# Copyright 2017-2018 Berk Onat, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import map
from builtins import range
from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction, ParsingContext
from nomadcore.simple_parser import SimpleMatcher as SM
from .AMBERDictionary import set_excludeList, set_includeList, get_updateDictionary, getList_MetaStrInDict, getDict_MetaStrInDict
from nomadcore.smart_parser.SmartParserDictionary import isMetaStrInDict
from contextlib import contextmanager
from . import AMBERCommon as AmberC
from . import trajectory_reader as TrajRead
import logging
import os
import re
import sys


############################################################
# This is the parser for the main file of AMBER (mdout).
############################################################

LOGGER = logging.getLogger("nomad.AMBERParser")

#PRINTABLE = re.compile(r"\W+")

@contextmanager
def open_section(parser, name):
    gid = parser.openSection(name)
    yield gid
    parser.closeSection(name, gid)

class AMBERParser(AmberC.AMBERParserBase):
    """Context for parsing AMBER main file.

    This class keeps tracks of several AMBER settings to adjust the parsing to them.
    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of writing and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self):
        # dictionary of energy values, which are tracked between SCF iterations and written after convergence
        self.totalEnergyList = {
                                'energy_electrostatic': None,
                                'energy_total_T0_per_atom': None,
                                'energy_free_per_atom': None,
                               }
        AmberC.AMBERParserBase.__init__(
            self, re_program_name=re.compile(r"\s*Amber$"))

        self.cachingLevelForMetaName = {
                               'x_amber_trajectory_file_detect': CachingLevel.Cache,
                               'x_amber_geometry_optimization_cdetect': CachingLevel.Cache,
                               'x_amber_mdin_finline': CachingLevel.Ignore,
                               'x_amber_mdin_wt': CachingLevel.Ignore,
                               'x_amber_section_input_output_files': CachingLevel.Ignore,
                               'x_amber_single_configuration_calculation_detect': CachingLevel.Cache,
                              }
        for name in self.metaInfoEnv.infoKinds:
            metaInfo = self.metaInfoEnv.infoKinds[name]
            if (name.startswith('x_amber_mdin_') and
                metaInfo.kindStr == "type_document_content" and
                ("x_amber_mdin_method" in metaInfo.superNames or
                 "x_amber_mdin_run" in metaInfo.superNames or
                 "x_amber_mdin_system" in metaInfo.superNames) or
                name.startswith('x_amber_parm_') and
                metaInfo.kindStr == "type_document_content" and
                ("x_amber_mdin_method" in metaInfo.superNames or
                 "x_amber_mdin_run" in metaInfo.superNames or
                 "x_amber_mdin_system" in metaInfo.superNames) or
                name.startswith('x_amber_mdin_file_') and
                metaInfo.kindStr == "type_document_content" and
                ("x_amber_section_input_output_files" in metaInfo.superNames) or
                name.startswith('x_amber_mdout_') and
                metaInfo.kindStr == "type_document_content" and
                ("x_amber_mdout_method" in metaInfo.superNames or
                 "x_amber_mdout_system" in metaInfo.superNames or
                 "x_amber_mdout_single_configuration_calculation" in metaInfo.superNames)
                or name.startswith('section_single_configuration_calculation')
               ):
                self.cachingLevelForMetaName[name] = CachingLevel.Cache
            if name in self.extraDict.keys():
                self.cachingLevelForMetaName[name] = CachingLevel.Ignore


    def initialize_values(self):
        """Initializes the values of certain variables.

        This allows a consistent setting and resetting of the variables,
        when the parsing starts and when a section_run closes.
        """
        self.secMethodGIndex = None
        self.secSystemGIndex = None
        self.secTopologyGIndex = None
        self.secSamplingGIndex = None
        self.secSingleGIndex = None
        self.secVDWGIndex = None
        self.secAtomType = None
        self.inputMethodIndex = None
        self.mainMethodIndex = None
        self.mainCalcIndex = None
        self.MD = True
        self.topologyDict = None
        self.topologyTable = None
        self.topologyBonds = None
        self.topology = None
        self.topologyFormat = None
        self.topologyFile = None
        self.trajectory = None
        self.trajectoryFormat = None
        self.trajectoryFile = None
        self.readChunk = 300
        self.boxlengths = None
        self.latticevectors = None
        self.atompositions = None
        # start with -1 since zeroth iteration is the initialization
        self.MDiter = -1
        self.singleConfCalcs = []
        self.minConverged = None
        self.parsedLogFile = False
        self.LogSuperContext = None
        self.forces_raw = []
        self.lastCalculationGIndex = None
        self.logFileName = None
        self.lastfInLine = None
        self.lastfInMatcher = None
        self.secOpen = open_section
        self.superP = self.parser.backend.superBackend

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts.

        Get compiled parser, filename and metadata.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        self.fName = fInName
        # save metadata
        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def peekline(self, parser):
        pos = parser.fIn.fIn.tell()
        line = parser.fIn.fIn.readline()
        parser.fIn.fIn.seek(pos)
        return line

    def topologyFileHandler(self, fileItem):
        if 'topology' in self.fileDict[fileItem].infoPurpose:
            topofile = self.fileDict[fileItem].fileName
            self.topologyFile = topofile
            self.trajectory = TrajRead.TrajectoryReader()
            self.trajectory.topofile = topofile
            self.topology = self.trajectory.load_topology()
            if self.topology is None:
                for fileFormat in self.fileDict[fileItem].fileFormat:
                    self.trajectory = TrajRead.TrajectoryReader()
                    self.trajectory.topofile = topofile
                    self.trajectory.topoformat = fileFormat
                    self.topology = self.trajectory.load_topology()
                    if self.topology is not None:
                        return fileFormat

    def trajectoryFileHandler(self, fileItem, topoformat):
        if 'trajectory' in self.fileDict[fileItem].infoPurpose:
            trajfile = self.fileDict[fileItem].fileName
            self.trajectoryFile = trajfile
            self.trajectory = TrajRead.TrajectoryReader()
            self.trajectory.trajfile = trajfile
            self.trajectory.topofile = self.topologyFile
            self.trajectory.topoformat = topoformat
            traj_loaded = self.trajectory.load()
            self.trajectory.trajchunk=self.readChunk
            self.atompositions = self.trajectory.iread()
            if self.atompositions is None:
                for fileFormat in self.fileDict[fileItem].fileFormat:
                    self.trajectory = TrajRead.TrajectoryReader()
                    self.trajectory.trajfile = trajfile
                    self.trajectory.topofile = self.topologyFile
                    self.trajectory.topoformat = topoformat
                    self.trajectory.trajformat = fileFormat
                    self.trajectory.trajchunk=self.readChunk
                    traj_loaded = self.trajectory.load()
                    self.atompositions = self.trajectory.iread()
                    if self.atompositions is not None:
                        self.topology = self.trajectory.get_topology()
                        self.MDiter += 1
                        return fileFormat
            else:
                self.MDiter += 1

    def initializeFileHandlers(self):
        # Files will be loaded using their extensions initially.
        # If this fails, the fileFormat lists will be used in loading process.
        topoformat = None
        trajformat = None
        for fileItem in self.fileDict:
            if (self.fileDict[fileItem].fileSupplied and
                self.fileDict[fileItem].activeInfo):
                # First, check topology file
                topoformat = self.topologyFileHandler(fileItem)
                # Second, check trajectory file
                trajformat = self.trajectoryFileHandler(fileItem, topoformat)
        if self.topology:
            self.topologyTable, self.topologyBonds = self.topology.to_dataframe()
            self.topologyDict = self.topologyTable.to_dict(orient='list')
        self.topologyFormat = topoformat
        self.trajectoryFormat = trajformat
        #if trajformat or topoformat:
        #    return True
        #else:
        #    return False

    def parameter_file_name(self, itemdict):
        """ Function to generate data for parameter files list
        """
        working_dir_name = os.path.dirname(os.path.abspath(self.fName))
        parmmeta = isMetaStrInDict("PARM",self.fileDict)
        filename = []
        if parmmeta is not None:
            if self.fileDict[parmmeta].value is not None:
                filename.append(self.fileDict[parmmeta].value)
        if filename:
            return False, filename, itemdict
        else:
            return False, None, itemdict

    def topology_to_dictionary(self):
        """ This function generates self.topologyDict dictionary
            and converts/stores all the topology information
            according to the meta info definitions
        """
        newDict = {}
        return newDict

    def topology_num_topo_mol(self, itemdict):
        """ Function to generate data for number_of_topology_molecules
        """
        if self.topology:
            residueList = self.topologyDict["resSeq"]
            return False, residueList[len(residueList)-1], itemdict
        else:
            return False, None, itemdict

    def topology_system_name(self, itemdict):
        """ Function to generate data for system_name
        """
        system_name = self.fName.split('.')[-1]
        if self.topology:
            residueList = ','.join(list(set(self.topologyDict["resName"])))
        return False, system_name, itemdict

    def topology_atom_to_mol(self, itemdict):
        """ Function to generate data for atom_to_molecule
        """
        if self.topology:
            residueList = self.topologyDict["resSeq"]
            atomIndex = np.arange(len(residueList))
            atom_to_residue = np.zeros((len(residueList), 2), dtype=int)
            atom_to_residue[:,0] = atomIndex+1
            atom_to_residue[:,1] = np.array(residueList)+1
            return False, atom_to_residue, itemdict
        else:
            return False, None, itemdict

    def topology_atom_type_and_interactions(self, backend, gIndex):
        """ Function to generate data for atom_to_molecule
        """
        sO = open_section
        supB = backend.superBackend
        if self.topology:
            atom_type_list=list(set(self.topologyDict["name"]))
            atom_type_dict = {}
            massesDict = {}
            elementDict = {}
            radiusDict = {}
            for ielm in range(len(atom_type_list)):
                elm = atom_type_list[ielm]
                atom_type_dict.update({elm : ielm+1})
                for atom in self.topology.atoms:
                    if elm == atom.name:
                        massesDict.update({atom.name : atom.element.mass})
                        elementDict.update({atom.name : atom.element.symbol})
                        radiusDict.update({atom.name : atom.element.radius})

            massesList = list(massesDict.values())
            elementList = list(elementDict.values())
            radiusList = list(radiusDict.values())

            interNum = 0
            interDict = {}
            interTypeDict = {}
            for ielm in range(len(atom_type_list)-1):
                for jelm in range(ielm+1, len(atom_type_list)):
                    aelm = atom_type_list[ielm]
                    belm = atom_type_list[jelm]
                    bondList = []
                    typeList = []
                    bondid = 0
                    noninter = True
                    for bond in self.topology.bonds:
                        molname1, molatom1 = str(bond[0]).split('-')
                        molname2, molatom2 = str(bond[1]).split('-')
                        if((aelm == str(molatom1) and belm == str(molatom2)) or
                            (aelm == str(molatom2) and belm == str(molatom1))):
                            bondList.append(list(self.topologyBonds[bondid]))
                            interDict.update({interNum : bondList})
                            if noninter:
                                noninter = False
                                typeList.extend(list([
                                    atom_type_dict[aelm],
                                    atom_type_dict[belm]
                                    ]))
                                interTypeDict.update({interNum : typeList})
                                interNum += 1
                        bondid += 1

            for elm in range(len(atom_type_list)):
                with sO(supB, 'section_atom_type'):
                    ### !!! ----------------------------------------------- !!!
                    ### !!! Need to check the definition of atom type name.
                    ### !!! Which one is better? C1, CA, C:1 or C !!!
                    ### !!! ----------------------------------------------- !!!
                    # Atom name? Atom element?
                    supB.addValue('atom_type_name', str(atom_type_list[elm]) + ':' + str(elm+1))
                    # Atomic mass
                    supB.addValue('atom_type_mass', massesList[elm])
                    # Atomic element/symbol
                    supB.addValue('x_amber_atom_type_element', elementList[elm])
                    # Atomic van der Waals radius
                    supB.addValue('x_amber_atom_type_radius', radiusList[elm])
                    # Atomic charge
                    #self.superP.addValue('atom_type_charge', atom_charge)
                    pass

            for inum in range(interNum):
                with sO(supB, 'section_interaction'):
                    # atom indexes of bound pairs for a specific atom type
                    supB.addArrayValues('interaction_atoms', np.asarray(interDict[inum]))
                    # number of bonds for this type
                    supB.addValue('number_of_interactions', len(interDict[inum]))
                    # number of atoms involved (2 for covalent bonds)
                    supB.addValue('number_of_atoms_per_interaction', len(interDict[inum][0]))

                    #if bondFunctional:
                    #    self.superP.addValue('interaction_kind', bondFunctional)  # functional form of the interaction

                    # this points to the relative section_atom_type
                    supB.addArrayValues('x_amber_interaction_atom_to_atom_type_ref',
                            np.asarray(interTypeDict[inum]))
                    # interaction parameters for the functional
                    #self.superP.addValue('interaction_parameters', bondParameters)


#            for i in range(len(moleculeTypeInfo)):
#
#                with o(p, 'section_molecule_type'):
#                    # gindex = 0
#                    p.addValue('molecule_type_name', 'molecule'+'_'+str(moleculeTypeInfo[i][0]))
#                    p.addValue('number_of_atoms_in_molecule', len(moleculeTypeInfo[i][1]))
#
#                    p.addArrayValues('atom_in_molecule_to_atom_type_ref', np.asarray([x-1 for x in moleculeTypeInfo[i][1]]))
#
#
#            residueList = self.topologyDict["resSeq"]
#            atomIndex = np.arange(len(residueList))
#            atom_to_residue = np.zeros((len(residueList), 2), dtype=int)
#            atom_to_residue[:,0] = atomIndex
#            atom_to_residue[:,1] = residueList
#            return atom_to_residue
#        else:
#            return None

#    mol_to_mol_bond = []
#    mol_to_mol_bond_num = []
#    atom_in_mol_bond = []
#    atom_in_mol_bond_num = []
#    index=0
#    print(len([res for res in mytopology.bonds]))
#    for res in mytopology.bonds:
#        molname1, molatom1 = str(res[0]).split('-')
#        molname2, molatom2 = str(res[1]).split('-')
#        if molname1 in molname2:
#            atom_in_mol_bond.append(res)
#            atom_in_mol_bond_num.append(bonds[index])
#        else:
#            mol_to_mol_bond.append(res)
#            mol_to_mol_bond_num.append(bonds[index])
#        index += 1
#    print(mol_to_mol_bond)
#    print(np.array(mol_to_mol_bond_num))

    def onClose_section_run(self, backend, gIndex, section):
        """Trigger called when section_run is closed.

        Write convergence of geometry optimization.
        Write the keywords from control parametres and the Amber output from the parsed log output, which belong to settings_run.
        Write the last occurrence of a keyword/setting, i.e. [-1], since Amber uses the last occurrence of a keyword.
        Variables are reset to ensure clean start for new run.

        ATTENTION
        backend.superBackend is used here instead of only the backend to write the JSON values,
        since this allows to bybass the caching setting which was used to collect the values for processing.
        However, this also bypasses the checking of validity of the metadata name by the backend.
        The scala part will check the validity nevertheless. (Good to know :)
        """
        # write trajectory
        #valuesDict = section.simpleValues

        frameSequenceGIndex = backend.openSection("section_frame_sequence")
        self.metaStorage.updateBackend(backend,
                startsection=['section_frame_sequence'],
                autoopenclose=False)
        backend.addValue("frame_sequence_to_sampling_ref", self.secSamplingGIndex)
        backend.addArrayValues("frame_sequence_local_frames_ref", np.asarray(self.singleConfCalcs))
        backend.closeSection("section_frame_sequence", frameSequenceGIndex)

        # reset all variables
        self.initialize_values()

    def onClose_x_amber_section_input_output_files(self, backend, gIndex, section):
        """Trigger called when x_amber_section_input_output_files is closed.

        Determine whether topology, trajectory and input coordinate files are
        supplied to the parser

        Initiates topology and trajectory file handles.

        Captures topology, atomic positions, atom labels, lattice vectors and
        stores them before section_system and
        section_single_configuration_calculation are encountered.
        """
        # Checking whether topology, input
        # coordinates and trajectory files exist
        atLeastOneFileExist = False
        working_dir_name = os.path.dirname(os.path.abspath(self.fName))
        for k,v in section.simpleValues.items():
            if k.startswith('x_amber_mdin_file'):
                file_name = os.path.normpath(os.path.join(working_dir_name, v[-1]))
                self.fileDict[k].fileSupplied = os.path.isfile(file_name)
                if self.fileDict[k].fileSupplied:
                    self.fileDict[k].fileName = file_name
                    if self.fileDict[k].activeInfo:
                        self.fileDict[k].value = v[-1]
                        atLeastOneFileExist = True
        if atLeastOneFileExist:
            self.initializeFileHandlers()

    def onOpen_section_method(self, backend, gIndex, section):
        # keep track of the latest method section
        self.secMethodGIndex = gIndex
        if self.inputMethodIndex is None:
            self.inputMethodIndex = gIndex
        else:
            backend.openNonOverlappingSection("section_method_to_method_refs")
            backend.addValue("method_to_method_kind", "core_settings")
            backend.addValue("method_to_method_ref", self.inputMethodIndex)
            backend.closeNonOverlappingSection("section_method_to_method_refs")
        if self.mainMethodIndex is None:
            self.mainMethodIndex = gIndex

    def onClose_section_method(self, backend, gIndex, section):
        """Trigger called when section_method is closed.
        """
        # input method
        pass

    def onOpen_section_sampling_method(self, backend, gIndex, section):
        # keep track of the latest sampling description section
        self.secSamplingGIndex = gIndex

    def onClose_section_sampling_method(self, backend, gIndex, section):
        """Trigger called when section_sampling_method is closed.

        Writes sampling method details for minimization and molecular dynamics.
        """
        # check control keywords were found throguh dictionary support
        section_sampling_Dict = get_updateDictionary(self, 'sampling')
        updateDict = {
            'startSection' : [['section_sampling_method']],
            'dictionary' : section_sampling_Dict
            }
        #self.secSamplingGIndex = backend.openSection("section_sampling_method")
        self.metaStorage.update(updateDict)
        self.metaStorage.updateBackend(backend,
                startsection=['section_sampling_method'],
                autoopenclose=False)
        self.secRestrictGIndex = backend.openSection("section_restricted_uri")
        section_restrictions_Dict = get_updateDictionary(self, 'restrictions')
        updateDict = {
            'startSection' : [['section_restricted_uri']],
            'dictionary' : section_restrictions_Dict
            }
        self.metaStorageRestrict.update(updateDict)
        self.metaStorageRestrict.updateBackend(backend,
                startsection=['section_restricted_uri'],
                autoopenclose=False)
        backend.closeSection("section_restricted_uri", self.secRestrictGIndex)

    def onOpen_section_topology(self, backend, gIndex, section):
        # keep track of the latest topology description section
        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            self.secTopologyGIndex = backend.superBackend.openSection("section_topology")
        else:
            self.secTopologyGIndex = gIndex

    def onClose_section_topology(self, backend, gIndex, section):
        """Trigger called when section_topology is closed.
        """
        section_topology_Dict = get_updateDictionary(self, 'topology')
        updateDict = {
            'startSection' : [
                ['section_topology']],
            'muteSections' : [['section_sampling_method']],
            'dictionary' : section_topology_Dict
            }
        self.metaStorage.update(updateDict)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=['section_topology'],
                autoopenclose=False)
        self.topology_atom_type_and_interactions(backend, gIndex)

        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            backend.superBackend.closeSection("section_topology", self.secTopologyGIndex)

    def onOpen_section_system(self, backend, gIndex, section):
        # keep track of the latest system description section
        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            self.secSystemGIndex = backend.superBackend.openSection("section_system")
        else:
            self.secSystemGIndex = gIndex

    def onClose_section_system(self, backend, gIndex, section):
        """Trigger called when section_system is closed.

        Writes atomic positions, atom labels and lattice vectors.
        """
        # Our special recipe for the confused backend and more
        # See our other recipes at the back of this parser: MetaInfoStorage! :)
        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            SloppyBackend = backend.superBackend
        else:
            SloppyBackend = backend

        if self.topology:
            if (self.secTopologyGIndex is None or
                (self.secTopologyGIndex == -1 or
                self.secTopologyGIndex == "-1")):
                self.onOpen_section_topology(backend, None, None)
                self.onClose_section_topology(backend, None, None)
            SloppyBackend.addValue("topology_ref", self.secTopologyGIndex)

        if self.atompositions is not None:
            self.trajRefSingleConfigurationCalculation = gIndex
            unit_cell = np.asarray(self.metaStorage.convertUnits(
                self.atompositions.unitcell_vectors[0], "Angstrom", self.unitDict))
            SloppyBackend.addArrayValues('simulation_cell', unit_cell)
            SloppyBackend.addArrayValues('lattice_vectors', unit_cell)
            SloppyBackend.addArrayValues('atom_positions', np.transpose(np.asarray(
                self.metaStorage.convertUnits(self.atompositions.xyz[0], "Angstrom", self.unitDict))))
            if self.topology is not None:
                atom_labels = self.topologyDict["element"]
                SloppyBackend.addArrayValues('atom_labels', np.asarray(atom_labels))
                pass

            # Read the next step at trajectory in advance
            # If iread returns None, it will be the last step
            try:
                self.atompositions = self.trajectory.iread()
                self.MDiter += 1
            except AttributeError:
                self.atompositions = None
                pass

            #if atom_vel:
            # need to transpose array since its shape is [number_of_atoms,3] in the metadata
            #backend.addArrayValues('atom_velocities', np.transpose(np.asarray(atom_vel)))

        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            SloppyBackend.closeSection("section_system", self.secSystemGIndex)

    def onOpen_section_single_configuration_calculation(self, backend, gIndex, section):
        # write the references to section_method and section_system
        backend.addValue('single_configuration_to_calculation_method_ref', self.secMethodGIndex)
        backend.addValue('single_configuration_calculation_to_system_ref', self.secSystemGIndex)
        self.singleConfCalcs.append(gIndex)
        self.secSingleGIndex = backend.superBackend.openSection("section_single_configuration_calculation")

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        """Trigger called when section_single_configuration_calculation is closed.

        Write number of steps in MD or Minimization.
        Check for convergence of geometry optimization.
        Write energy values at MD and with error in Minimization.
        Write reference to section_method and section_system
        """

        self.lastCalculationGIndex = gIndex

        section_frameseq_Dict = get_updateDictionary(self, 'frameseq')
        updateFrameDict = {
            'startSection' : [
                ['section_frame_sequence']],
            'muteSections' : [['section_sampling_method']],
            'dictionary' : section_frameseq_Dict
            }
        self.metaStorage.update(updateFrameDict)
        section_singlevdw_Dict = get_updateDictionary(self, 'singlevdw')
        updateDictVDW = {
            'startSection' : [
                ['section_energy_van_der_Waals']],
            'muteSections' : [['section_sampling_method']],
            'dictionary' : section_singlevdw_Dict
            }
        self.secVDWGIndex = backend.superBackend.openSection("section_energy_van_der_Waals")
        self.metaStorage.update(updateDictVDW)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=['section_energy_van_der_Waals'],
                autoopenclose=False)
        backend.superBackend.closeSection("section_energy_van_der_Waals", self.secVDWGIndex)
        section_singlecalc_Dict = get_updateDictionary(self, 'singleconfcalc')
        updateDict = {
            'startSection' : [
                ['section_single_configuration_calculation']],
            'muteSections' : [['section_sampling_method']],
            'dictionary' : section_singlecalc_Dict
            }
        self.metaStorage.update(updateDict)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=['section_single_configuration_calculation'],
                autoopenclose=False)
        if(self.topology is not None or
           self.atompositions is not None):
            self.onOpen_section_system(backend, None, None)
            self.onClose_section_system(backend, None, None)
        backend.superBackend.closeSection("section_single_configuration_calculation", self.secSingleGIndex)

        # write number of Minimization steps
        #backend.addValue('number_of_steps', self.minStepNr)
#        # write Minimization convergence and reset
#        backend.addValue('single_configuration_calculation_converged', self.minConvergence)
#        self.minConvergence = False
#        # start with -1 since zeroth iteration is the initialization
#        self.minStepNr = -1
#        # write converged energy/thermodynamic values
#        # write forces
#        forces_free = []
#        for i in ['x', 'y', 'z']:
#            fi = section['x_amber_atom_forces_free_' + i]
#            if fi is not None:
#                forces_free.append(fi)
#        if forces_free:
#            # need to transpose array since its shape is [number_of_atoms,3] in the metadata
#            backend.addArrayValues('atom_forces_free', np.transpose(np.asarray(forces_free)))
#        if self.forces_raw:
#            # need to transpose array since its shape is [number_of_atoms,3] in the metadata
#            backend.addArrayValues('atom_forces_free_raw', np.transpose(np.asarray(self.forces_raw)))

    def setStartingPointCalculation(self, parser):
        backend = parser.backend
        backend.openSection('section_calculation_to_calculation_refs')
        if self.lastCalculationGIndex:
            backend.addValue('calculation_to_calculation_ref', self.lastCalculationGIndex)
#        backend.addValue('calculation_to_calculation_kind', 'pertubative GW')
            backend.closeSection('section_calculation_to_calculation_refs')
        return None

    def check_namelist_store(self, parser, lastLine, stopOnMatchRe, quitOnMatchRe,
            metaNameStart, matchNameList, matchNameDict, onlyCaseSensitive, stopOnFirstLine):
        stopOnMatch = False
        if stopOnMatchRe.findall(lastLine):
            stopOnMatch = True
            if self.firstLine==0:
                if stopOnFirstLine:
                    stopOnMatch = True
                else:
                    stopOnMatch = False
        if quitOnMatchRe is not None:
            if quitOnMatchRe.findall(lastLine):
                stopOnMatch = True
        if stopOnMatch:
            return True
        else:
            # If there is at least one namelist in the line,
            # search for all others in the dictionary.
            if self.MD is not True:
                newLine = parser.fIn.readline()
                lastLine = ' = '.join([ "%s" % str(line) for line in zip(lastLine, newLine)])
            for cName, key in getDict_MetaStrInDict(matchNameDict).items():
                reDict={key:value for value in
                        re.compile(r"(?:\s%s|^%s|,%s)\s*=\s*(?:'|\")?(?P<%s>[\-+0-9.a-zA-Z:]+)(?:'|\")?\s*,?"
                        % (cName, cName, cName, key)).findall(lastLine)}
                if onlyCaseSensitive is not True:
                    reDict.update({key:value for value in
                                   re.compile(r"(?:\s%s|^%s|,%s)\s*=\s*(?:'|\")?(?P<%s>[\-+0-9.a-zA-Z:]+)(?:'|\")?\s*,?"
                                   % (cName.upper(), cName.upper(), cName.upper(), key)).findall(lastLine)})
                if reDict:
                    for k,v in reDict.items():
                        if k == key:
                            if k in list(parser.lastMatch.keys()):
                                parser.lastMatch[k]=v
                            else:
                                matchNameDict[k].value=v
                                matchNameDict[k].activeInfo=True
            return False

    def adHoc_read_namelist_stop_parsing(self, parser, stopOnMatchStr, quitOnMatchStr,
            metaNameStart, matchNameList, matchNameDict, onlyCaseSensitive, stopOnFirstLine):
        lastLine = parser.fIn.fInLine
        self.firstLine = 0
        # Check the captured line has Fortran namelist variables and store them.
        # Continue search and store until the line matches with stopOnMatch.
        stopOnMatchRe = re.compile(stopOnMatchStr)
        quitOnMatchRe = None
        if quitOnMatchStr is not None:
            quitOnMatchRe = re.compile(quitOnMatchStr)
        if self.check_namelist_store(parser, lastLine,
                stopOnMatchRe, quitOnMatchRe,
                metaNameStart, matchNameList,
                matchNameDict, onlyCaseSensitive,
                stopOnFirstLine) is not True:
            while True:
                lastLine = self.peekline(parser)
                self.firstLine += 1
                if not lastLine:
                    break
                else:
                    # Matched with stopOnMatch. Discarding the line and return SimpleMatcher context.
                    # Can this line be discarded since it is the end of line for input control
                    # variables or end of namelist ?
                    if self.check_namelist_store(parser, lastLine,
                            stopOnMatchRe, quitOnMatchRe,
                            metaNameStart, matchNameList,
                            matchNameDict, onlyCaseSensitive,
                            stopOnFirstLine):
                        break
                    else:
                        lastLine = parser.fIn.readline()

    def build_mdinKeywordsSimpleMatchers(self):
        cntrlNameList=getList_MetaStrInDict(self.cntrlDict)
        ewaldNameList=getList_MetaStrInDict(self.ewaldDict)
        qmmmNameList=getList_MetaStrInDict(self.qmmmDict)
        wtNameList=getList_MetaStrInDict(self.wtDict)
        return [
            SM(name="cntrl",
               startReStr=r"\s*&cntrl",
               endReStr=r"(?:(?:^/|\s*/)|&end|&END)\s*$",
               subMatchers=[
                   SM(startReStr=(r"\s*(?:" +
                      '|'.join(["%s" % (cName) for cName in cntrlNameList]) +
                      "AMBER)\s*=\s*(?:'|\")?(?P<x_amber_mdin_finline>[\-+0-9.:a-zA-Z]+)(?:'|\")?\s*,?"),
                      coverageIgnore=True,
                      adHoc=lambda p:
                      self.adHoc_read_namelist_stop_parsing(p,
                      stopOnMatchStr=r"(?:(?:^/|\s*/)|&end|&END)\s*$",
                      quitOnMatchStr=None,
                      metaNameStart="x_amber_mdin_",
                      matchNameList=cntrlNameList,
                      matchNameDict=self.cntrlDict,
                      onlyCaseSensitive=True,
                      stopOnFirstLine=True)
                      )
               ]),
            SM(name="wt",
                startReStr=r"(?:\s*&wt|^&wt)",
                endReStr=r"\s*&wt\s*(?:type|TYPE)\s*=\s*(?:\'|\")(?:end|END)(?:\'|\")\s*\/\s*",
                forwardMatch=True,
                subMatchers=[
                    SM(startReStr=r"(?:\s*&wt|^&wt)\s*(?:type|TYPE)\s*=" +
                       "(?:'|\")?(?P<x_amber_mdin_wt>[-+0-9.:a-zA-Z]+)(?:'|\")?",
                       repeats=True,
                       adHoc=lambda p:
                       self.adHoc_read_namelist_stop_parsing(p,
                       stopOnMatchStr=r"\s*&wt\s*(?:type|TYPE)\s*=\s*(?:\'|\")(?:end|END)(?:\'|\")\s*\/\s*",
                       quitOnMatchStr=None,
                       metaNameStart="x_amber_mdin_",
                       matchNameList=wtNameList,
                       matchNameDict=self.wtDict,
                       onlyCaseSensitive=False,
                       stopOnFirstLine=True)
                       )
                ]),
            SM(name="ewald",
               startReStr=r"\s*&ewald",
               endReStr=r"(?:(?:^/|\s*/)|&end|&END)\s*$",
               subMatchers=[
                   SM(startReStr=(r"\s*(?:" +
                      '|'.join(["%s" % (cName) for cName in ewaldNameList]) +
                      "AMBER)\s*=\s*(?:'|\")?(?P<x_amber_mdin_finline>[\-+0-9.:a-zA-Z]+)(?:'|\")?\s*,?"),
                      coverageIgnore=True,
                      adHoc=lambda p:
                      self.adHoc_read_namelist_stop_parsing(p,
                      stopOnMatchStr=r"(?:(?:^/|\s*/)|&end|&END)\s*$",
                      quitOnMatchStr=None,
                      metaNameStart="x_amber_mdin_",
                      matchNameList=ewaldNameList,
                      matchNameDict=self.ewaldDict,
                      onlyCaseSensitive=True,
                      stopOnFirstLine=True)
                      )
               ]),
            SM(name="qmmm",
               startReStr=r"\s*&qmmm",
               endReStr=r"(?:(?:^/|\s*/)|&end|&END)\s*$",
               subMatchers=[
                   SM(startReStr=(r"\s*(?:" +
                      '|'.join(["%s" % (cName) for cName in qmmmNameList]) +
                      "AMBER)\s*=\s*(?:'|\")?(?P<x_amber_mdin_finline>[\-+0-9.a-zA-Z:]+)(?:'|\")?\s*,?"),
                      coverageIgnore=True,
                      adHoc=lambda p:
                      self.adHoc_read_namelist_stop_parsing(p,
                      stopOnMatchStr=r"(?:(?:^/|\s*/)|&end|&END)\s*$",
                      quitOnMatchStr=None,
                      metaNameStart="x_amber_mdin_",
                      matchNameList=qmmmNameList,
                      matchNameDict=self.qmmmDict,
                      onlyCaseSensitive=True,
                      stopOnFirstLine=True)
                      )
               ])
            ]

    def build_parmKeywordsSimpleMatchers(self):
        parmNameList=getList_MetaStrInDict(self.parmDict)
        return [
            SM(name="parm",
               startReStr=r"\|\s*Version\s*=\s*(?P<x_amber_parm_file_version>[0-9.eEdD]+)" +
                           "\s*Date\s*=\s*(?P<x_amber_parm_file_date>[0-9a-zA-Z\/]+)" +
                           "\s*Time\s*=\s*(?P<x_amber_parm_file_time>[0-9a-zA-Z:]+)",
               endReStr=r"\|\s*Memory\s*Use\s*Allocated\s*",
               subMatchers=[
                   SM(startReStr=(r"\s*(?:" +
                      '|'.join(["%s" % (cName) for cName in parmNameList]) +
                      "AMBER)\s*=\s*(?:'|\")?(?P<x_amber_mdin_finline>[\-+0-9.:a-zA-Z]+)(?:'|\")?\s*,?"),
                      coverageIgnore=True,
                      adHoc=lambda p:
                      self.adHoc_read_namelist_stop_parsing(p,
                      stopOnMatchStr=r"\|\s*Memory\s*Use\s*Allocated\s*",
                      quitOnMatchStr=None,
                      metaNameStart="x_amber_parm_",
                      matchNameList=parmNameList,
                      matchNameDict=self.parmDict,
                      onlyCaseSensitive=True,
                      stopOnFirstLine=True)
                      )
               ])
            ]

    def build_mdoutKeywordsSimpleMatchers(self):
        newDict = self.cntrlDict
        newDict.update(self.qmmmDict)
        mdoutNameList = getList_MetaStrInDict(newDict)
        return [
            SM(name="mdout",
               startReStr=r"\s*General\s*flags:",
               endReStr=r"\s*--{5}--*\s*3\.\s*ATOMIC\s*COORDINATES\s*",
               subMatchers=[
                   SM(startReStr=(r"\s*(?:" +
                      '|'.join(["%s" % (cName) for cName in mdoutNameList]) +
                      "AMBER)\s*=\s*(?:'|\")?(?P<x_amber_mdin_finline>[\-+0-9.:a-zA-Z]+)(?:'|\")?\s*,?"),
                      coverageIgnore=True,
                      adHoc=lambda p:
                      self.adHoc_read_namelist_stop_parsing(p,
                      stopOnMatchStr=r"\s*3\.\s*ATOMIC\s*COORDINATES",
                      quitOnMatchStr=None,
                      metaNameStart="x_amber_mdin_",
                      matchNameList=mdoutNameList,
                      matchNameDict=newDict,
                      onlyCaseSensitive=True,
                      stopOnFirstLine=True)
                      )
               ])
            ]

    def build_fileNameListGroupSubMatcher(self):
        """Builds the Sub Matchers for the Main Parser
        """
        return [SM(r"\|\s*%s:\s*(?P<%s>[0-9a-zA-Z_./\-]+)" %
                (fileNL.matchStr, fileNL.metaHeader + '_' + fileNL.metaNameTag + '_' + fileNL.metaName)) for fileNL in self.fileDict.values()]


    def build_subMatchers(self):
        """Builds the sub matchers to parse the main file of AMBER.
        """
        mdinKeywordsSimpleMatchers = self.build_mdinKeywordsSimpleMatchers()
        parmKeywordsSimpleMatchers = self.build_parmKeywordsSimpleMatchers()
        mdoutKeywordsSimpleMatchers = self.build_mdoutKeywordsSimpleMatchers()
        fileNameListGroupSubMatcher = self.build_fileNameListGroupSubMatcher()

        ########################################
        # submatcher for mdin
        mdinSubMatcher = SM(name='mdinKeywords',
            startReStr=r"\s*Here\s*is\s*the\s*input\s*file:\s*",
            endReStr=r"\s*--{5}--*\s*1\.\s*RESOURCE\s*USE:",
            subFlags=SM.SubFlags.Unordered,
            subMatchers=[
                SM(r"(?:(?P<x_amber_mdin_header>[0-9a-zA-Z]+)?)\s*$",
                    coverageIgnore=True,
                    repeats=False),
                ] + mdinKeywordsSimpleMatchers
            )

        ########################################
        # submatcher for parm
        parmSubMatcher = SM(name='parmKeywords',
            startReStr=r"\s*--{5}--*\s*|\s*Flags:",
            endReStr=r"\s*--{5}--*\s*2\.\s*CONTROL\s*",
            forwardMatch=True,
            #subFlags=SM.SubFlags.Unordered,
            subMatchers=[
                SM(r"\|\s*Flags\s*:\s*(?P<x_amber_parm_flags>[0-9a-zA-Z]+)?\s*$"),
                SM(r"\s*getting\s*new\s*box\s*info\s*from\s*(?:bottom\s*of\s*|netcdf)(?P<x_amber_mdin_finline>[0-9a-zA-Z]+)\s*(file)?\s*",
                   coverageIgnore=True),
                SM(r"\|(\s*NetCDF)?\s*(?P<x_amber_parm_box_info>[_0-9a-zA-Z]+)\s*(?:ntb=[0-9]\s*and\s*igb=[0-9])?:?" +
                    "(?:(?:box|Box)info\s*found|Setting\s*up\s*nonperiodic\s*simulation)\s*"),
                SM(r"\|\s*Largest\s*sphere\s*to\s*fit\s*in\s*unit\s*cell\s*has\s*radius\s*=\s*" +
                    "(?P<x_amber_parm_unitcell_radius>[0-9.eEdD]+)\s*"),
                SM(r"\|\s*(?P<x_amber_parm_file_format>[0-9.eEdD]+)" +
                    "\s*format\s*PARM\s*file\s*being\s*parsed\s*\.")
                ] + parmKeywordsSimpleMatchers + [
                SM(r"\|\s*Real\s*(?P<x_amber_mdin_finline>[0-9.eEdD]+)\s*", coverageIgnore=True),
                SM(r"\|\s*Hollerith\s*(?P<x_amber_mdin_finline>[0-9.eEdD]+)\s*", coverageIgnore=True),
                SM(r"\|\s*Integer\s*(?P<x_amber_mdin_finline>[0-9.eEdD]+)\s*", coverageIgnore=True),
                SM(r"\|\s*Max\s*Pairs\s*(?P<x_amber_mdin_finline>[0-9.eEdD]+)\s*", coverageIgnore=True),
                SM(r"\|\s*nblistReal\s*(?P<x_amber_mdin_finline>[0-9.eEdD]+)\s*", coverageIgnore=True),
                SM(r"\|\s*nblist\s*Int\s*(?P<x_amber_mdin_finline>[0-9.eEdD]+)\s*", coverageIgnore=True),
                SM(r"\|\s*Total\s*(?P<x_amber_parm_total_memory>[0-9.eEdD]+)\s*")
                ]
            )

        ########################################
        # submatcher for mdout
        mdoutSubMatcher = SM(name='mdoutKeywords',
            startReStr=r"\s*General\s*flags:",
            forwardMatch=True,
            #subFlags=SM.SubFlags.Unordered,
            subMatchers=mdoutKeywordsSimpleMatchers
            )

        ########################################
        # submatcher for MD
        mddataNameList=getList_MetaStrInDict(self.mddataDict)
        MDSubMatcher = SM(name='MDStep',
            startReStr=r"\s*(?:NSTEP\s*=|NSTEP\s*ENERGY\s*RMS\s*)",
            #endReStr=r"\s*(?:FINAL\s*RESULTS|A\sV\sE\sR\sA\sG\sE\sS\s*O\sV\sE\sR)",
            forwardMatch=True,
            subMatchers=[
                   SM(startReStr=(r"\s*(?:(?:" +
                   '|'.join(["%s" % (cName) for cName in mddataNameList]) +
                   "AMBER)\s*=\s*|NSTEP\s*ENERGY\s*RMS\s*)(?:'|\")?" +
                   "(?P<x_amber_mdin_finline>[\-+0-9.:a-zA-Z]+)(?:'|\")?\s*,?"),
                   #endReStr=r"\s*(?:FINAL\s*RESULTS|A\sV\sE\sR\sA\sG\sE\sS\s*O\sV\sE\sR)",
                   coverageIgnore=True,
                   adHoc=lambda p:
                   self.adHoc_read_namelist_stop_parsing(p,
                   stopOnMatchStr=r"\s*(?:NSTEP\s*=|NSTEP\s*ENERGY\s*RMS\s*)",
                   quitOnMatchStr=r"\s*(?:FINAL\s*RESULTS|A\sV\sE\sR\sA\sG\sE\sS\s*O\sV\sE\sR)",
                   metaNameStart="x_amber_mdout_",
                   matchNameList=mddataNameList,
                   matchNameDict=self.mddataDict,
                   onlyCaseSensitive=True,
                   stopOnFirstLine=False)
                   )
            ])

        ########################################
        # return main Parser
        return [
            SM(name='NewRun',
                startReStr=r"\s*Amber\s*[0-9]+\s*(?:SANDER|PMEMD)",
                endReStr=r"\s*wallclock\(\) was called\s*[0-9]+\s*times",
                repeats=True,
                required=True,
                forwardMatch=True,
                sections=['section_run'],
                fixedStartValues={'program_name': 'Amber'},
                subMatchers=[
                    # header specifing version, compilation info, task assignment
                    SM(r"\s*Amber\s*(?P<program_version>[0-9a-zA-Z]+)\s*(?P<x_amber_program_module>[0-9a-zA-Z]+)\s*(?P<x_amber_program_version_date>[0-9a-zA-Z]+)"),
                    SM(name='run_date', startReStr=r"\|\s*Run\s*on\s*(?P<x_amber_program_execution_date>[0-9a-zA-Z\/]+)\s*at\s*(?P<x_amber_program_execution_time>[0-9a-zA-Z:]+)"),
                    SM(name='run_path', startReStr=r"\|\s*Executable\s*path\s*:\s*(?P<x_amber_program_execution_path>[0-9a-zA-Z\/]+)"),
                    SM(name='work_path', startReStr=r"\|\s*Working\s*directory\s*:\s*(?P<x_amber_program_working_path>[0-9a-zA-Z\/]+)"),
                    SM(name='run_host', startReStr=r"\|\s*Hostname\s*:\s*(?P<x_amber_program_execution_host>[0-9a-zA-Z\/]+)"),
                    SM(name='FileNameMatch',
                       startReStr=r"\s*File\s*Assignments:",
                       forwardMatch=True,
                       sections=['x_amber_section_input_output_files'],
                       subFlags=SM.SubFlags.Unordered,
                       subMatchers=fileNameListGroupSubMatcher
                       ), # END SectionMethod
                    SM(name='SectionMethod',
                       startReStr=r"\s*Here\s*is\s*the\s*input\s*file:",
                       forwardMatch=True,
                       sections=['section_sampling_method'],
                       subMatchers=[
                           # parse verbatim writeout of mdin file
                           mdinSubMatcher,
                           # parse summary writeout of topology parameters
                           parmSubMatcher,
                           # parse control settings writeout of Amber
                           mdoutSubMatcher
                       ]), # END SectionMethod
                    SM(name='SingleConfigurationCalculationWithSystemDescription',
                       startReStr=r"\s*4\.\s*RESULTS",
                       endReStr=r"\s*(?:FINAL\s*RESULTS|A\sV\sE\sR\sA\sG\sE\sS\s*O\sV\sE\sR)",
                       forwardMatch=True,
                       subMatchers=[
                           # the actual section for a single configuration calculation starts here
                           SM(name='SingleConfigurationCalculation',
                              startReStr=r"\s*(?:NSTEP\s*=|NSTEP\s*ENERGY\s*RMS\s*)",
                              #endReStr=r"\s*(?:FINAL\s*RESULTS|A\sV\sE\sR\sA\sG\sE\sS\s*O\sV\sE\sR)",
                              repeats=True,
                              forwardMatch=True,
                              sections=['section_single_configuration_calculation'],
                              subMatchers=[
                                  # MD
                                  MDSubMatcher
                              ]) # END SingleConfigurationCalculation
                       ]), # END SingleConfigurationCalculationWithSystemDescription
                    SM(r"\s*5\.\s*TIMINGS\s*"),
                    # summary of computation
                    SM(name='ComputationTimings',
                       startReStr=r"\s*Computation timings:",
                       subMatchers=[
                           SM(r"\s*Run\s*done\s*at[0-9:]+"),
                       ]), # END Computation
                    SM(name='end_run', startReStr=r"\s*wallclock\(\)\s*was\s*called\s*[0-9]+\s*times")
                ]) # END NewRun
            ]


class AMBERParserInterface():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
        from unittest.mock import patch
        logging.info('amber parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("amber.nomadmetainfo.json")
        parserInfo = {'name': 'amber-parser', 'version': '1.0'}
        context = AMBERParser()
        context.coverageIgnore = re.compile(r"^(?:" + r"|".join(context.coverageIgnoreList) + r")$")
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription=context.mainFileDescription(),
                metaInfoEnv=None,
                parserInfo=parserInfo,
                cachingLevelForMetaName=context.cachingLevelForMetaName,
                superContext=context,
                superBackend=backend)
        return backend


if __name__ == "__main__":
    parser = AMBERParser()
    parser.parse()



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Common data, variables, decorators, models used throughout the API.
'''
from typing import Callable, IO, Set, Tuple, Iterable, Dict, Any
from flask_restplus import fields
from flask import request, make_response
import zipstream
from flask import stream_with_context, Response, g, abort
from urllib.parse import urlencode
import pprint
import io
import json

import sys
import os.path
import gzip
from functools import wraps

from nomad import search, config, datamodel
from nomad.app.optimade import filterparser
from nomad.app.common import RFC3339DateTime, rfc3339DateTime
from nomad.files import Restricted

from .api import api


if sys.version_info >= (3, 7):
    import zipfile
else:
    import zipfile37 as zipfile


metadata_model = api.model('MetaData', {
    'with_embargo': fields.Boolean(default=False, description='Data with embargo is only visible to the upload until the embargo period ended.'),
    'comment': fields.String(description='The comment are shown in the repository for each calculation.'),
    'references': fields.List(fields.String, descriptions='References allow to link calculations to external source, e.g. URLs.'),
    'coauthors': fields.List(fields.String, description='A list of co-authors given by user_id.'),
    'shared_with': fields.List(fields.String, description='A list of users to share calculations with given by user_id.'),
    '_upload_time': RFC3339DateTime(description='Overrride the upload time.'),
    '_uploader': fields.String(description='Override the uploader with the given user id.'),
    'datasets': fields.List(fields.String, description='A list of dataset ids.')
})

pagination_model = api.model('Pagination', {
    'total': fields.Integer(description='Number of total elements.'),
    'page': fields.Integer(description='Number of the current page, starting with 0.'),
    'per_page': fields.Integer(description='Number of elements per page.'),
    'order_by': fields.String(description='Sorting criterion.'),
    'order': fields.Integer(description='Sorting order -1 for descending, 1 for asceding.')
})
''' Model used in responses with pagination. '''

scroll_model = api.model('Scroll', {
    'scroll': fields.Boolean(default=False, description='Flag if scrolling is enables.'),
    'total': fields.Integer(default=0, description='The total amount of hits for the search.'),
    'scroll_id': fields.String(default=None, allow_null=True, description='The scroll_id that can be used to retrieve the next page.'),
    'size': fields.Integer(default=0, help='The size of the returned scroll page.')})
''' Model used in responses with scroll. '''

aggregation_model = api.model('Aggregation', {
    'after': fields.String(description='The after key for the current request.', allow_null=True),
    'total': fields.Integer(default=0, description='The total amount of hits for the search.'),
    'per_page': fields.Integer(default=0, help='The size of the requested page.', allow_null=True)})
''' Model used in responses with id aggregation. '''

query_model_fields = {
    qualified_name: quantity.flask_field
    for qualified_name, quantity in search.search_quantities.items()}

query_model_fields.update(**{
    'owner': fields.String(description='The group the calculations belong to.', allow_null=True, skip_none=True),
    'domain': fields.String(description='Specify the domain to search in: %s, default is ``%s``' % (
        ', '.join(['``%s``' % domain for domain in datamodel.domains]), config.meta.default_domain)),
    'from_time': fields.Raw(description='The minimum entry time.', allow_null=True, skip_none=True),
    'until_time': fields.Raw(description='The maximum entry time.', allow_null=True, skip_none=True)
})

query_model_fields.update(**{
    '$and': fields.List(fields.Raw, description=(
        'List of queries which must be present in search results.')),
    '$or': fields.List(fields.Raw, description=(
        'List of queries which should be present in search results.')),
    '$not': fields.List(fields.Raw, description=(
        'List of queries which must not be present in search results.')),
    '$lt': fields.Raw(description=(
        'Dict of quantiy name: value such that search results should have values '
        'less than value.')),
    '$lte': fields.Raw(description=(
        'Dict of quantiy name: value such that search results should have values '
        'less than or equal to value')),
    '$gt': fields.Raw(description=(
        'Dict of quantiy name: value such that search results should have values '
        'greater than value')),
    '$gte': fields.Raw(description=(
        'Dict of quantiy name: value such that search results should have values '
        'greater than or equal to value')),
})

query_model = api.model('Query', query_model_fields)

search_model_fields = {
    'query': fields.Nested(query_model, allow_null=True, skip_none=True),
    'pagination': fields.Nested(pagination_model, allow_null=True, skip_none=True),
    'scroll': fields.Nested(scroll_model, allow_null=True, skip_none=True),
    'aggregation': fields.Nested(aggregation_model, allow_null=True),
    'results': fields.List(fields.Raw(allow_null=True, skip_none=True), description=(
        'A list of search results. Each result is a dict with quantitie names as key and '
        'values as values'), allow_null=True, skip_none=True),
    'code': fields.Nested(api.model('Code', {
        'repo_url': fields.String(description=(
            'An encoded URL for the search query on the repo api.')),
        'python': fields.String(description=(
            'A piece of python code snippet which can be executed to reproduce the api result.')),
        'curl': fields.String(description=(
            'A curl command which can be executed to reproduce the api result.')),
        'clientlib': fields.String(description=(
            'A piece of python code which uses NOMAD\'s client library to access the archive.'))
    }), allow_null=True, skip_none=True)}

search_model = api.model('Search', search_model_fields)


def add_pagination_parameters(request_parser):
    ''' Add pagination parameters to Flask querystring parser. '''
    request_parser.add_argument(
        'page', type=int, help='The page, starting with 1.', location='args')
    request_parser.add_argument(
        'per_page', type=int, help='Desired calcs per page.', location='args')
    request_parser.add_argument(
        'order_by', type=str, help='The field to sort by.', location='args')
    request_parser.add_argument(
        'order', type=int, help='Use -1 for decending and 1 for acending order.', location='args')


request_parser = api.parser()
add_pagination_parameters(request_parser)
pagination_request_parser = request_parser.copy()


def add_scroll_parameters(request_parser):
    ''' Add scroll parameters to Flask querystring parser. '''
    request_parser.add_argument(
        'scroll', type=bool, help='Enable scrolling')
    request_parser.add_argument(
        'scroll_id', type=str, help='The id of the current scrolling window to use.')


def add_search_parameters(request_parser):
    ''' Add search parameters to Flask querystring parser. '''
    # more search parameters
    request_parser.add_argument(
        'domain', type=str,
        help='Specify the domain to search in: %s, default is ``%s``' % (
            ', '.join(['``%s``' % domain for domain in datamodel.domains]),
            config.meta.default_domain))
    request_parser.add_argument(
        'owner', type=str,
        help='Specify which calcs to return: ``visible``, ``public``, ``all``, ``user``, ``staging``, default is ``visible``')
    request_parser.add_argument(
        'from_time', type=lambda x: rfc3339DateTime.parse(x),
        help='A yyyy-MM-ddTHH:mm:ss (RFC3339) minimum entry time (e.g. upload time)')
    request_parser.add_argument(
        'until_time', type=lambda x: rfc3339DateTime.parse(x),
        help='A yyyy-MM-ddTHH:mm:ss (RFC3339) maximum entry time (e.g. upload time)')
    request_parser.add_argument(
        'dft.optimade', type=str,
        help='A search query in the optimade filter language.')
    request_parser.add_argument(
        'query', type=str,
        help='A json serialized structured search query (as used in POST reuquests).')

    # main search parameters
    for qualified_name, quantity in search.search_quantities.items():
        request_parser.add_argument(
            qualified_name, help=quantity.description, action=quantity.argparse_action)


_search_quantities = set(search.search_quantities.keys())


def apply_search_parameters(search_request: search.SearchRequest, args: Dict[str, Any]):
    '''
    Help that adds query relevant request args to the given SearchRequest.
    '''
    args = {key: value for key, value in args.items() if value is not None}

    # domain
    domain = args.get('domain')
    if domain is not None:
        search_request.domain(domain=domain)

    # owner
    owner = args.get('owner', 'visible')
    try:
        search_request.owner(
            owner,
            g.user.user_id if g.user is not None else None)
    except ValueError as e:
        abort(401, getattr(e, 'message', 'Invalid owner parameter: %s' % owner))
    except Exception as e:
        abort(400, getattr(e, 'message', 'Invalid owner parameter'))

    # time range
    from_time_str = args.get('from_time', None)
    until_time_str = args.get('until_time', None)

    try:
        from_time = rfc3339DateTime.parse(from_time_str) if from_time_str is not None else None
        until_time = rfc3339DateTime.parse(until_time_str) if until_time_str is not None else None
        search_request.time_range(start=from_time, end=until_time)
    except Exception:
        abort(400, 'bad datetime format')

    # optimade
    try:
        optimade = args.get('dft.optimade', None)
        if optimade is not None:
            q = filterparser.parse_filter(
                optimade, nomad_properties=domain, without_prefix=True)
            search_request.query(q)
    except filterparser.FilterException as e:
        abort(400, 'Could not parse optimade query: %s' % str(e))

    # search expression
    query_str = args.get('query', None)
    if query_str is not None:
        try:
            query = json.loads(query_str)
        except Exception as e:
            abort(400, 'Could not JSON parse query expression: %s' % str(e))

        try:
            search_request.query_expression(query)
        except Exception as e:
            abort(400, 'Invalid query expression: %s' % str(e))

    # search parameter
    search_request.search_parameters(**{
        key: value for key, value in args.items()
        if key in _search_quantities})


def calc_route(ns, prefix: str = ''):
    ''' A resource decorator for /<upload>/<calc> based routes. '''
    def decorator(func):
        ns.route('%s/<string:upload_id>/<string:calc_id>' % prefix)(
            api.doc(params={
                'upload_id': 'The unique id for the requested upload.',
                'calc_id': 'The unique id for the requested calculation.'
            })(func)
        )
    return decorator


def upload_route(ns, prefix: str = ''):
    ''' A resource decorator for /<upload> based routes. '''
    def decorator(func):
        ns.route('%s/<string:upload_id>' % prefix)(
            api.doc(params={
                'upload_id': 'The unique id for the requested upload.'
            })(func)
        )
    return decorator


def streamed_zipfile(
        files: Iterable[Tuple[str, str, Callable[[str], IO], Callable[[str], int]]],
        zipfile_name: str, compress: bool = False):
    '''
    Creates a response that streams the given files as a streamed zip file. Ensures that
    each given file is only streamed once, based on its filename in the resulting zipfile.

    Arguments:
        files: An iterable of tuples with the filename to be used in the resulting zipfile,
            an file id within the upload, a callable that gives an binary IO object for the
            file id, and a callable that gives the file size for the file id.
        zipfile_name: A name that will be used in the content disposition attachment
            used as an HTTP respone.
        compress: Uses compression. Default is stored only.
    '''

    streamed_files: Set[str] = set()

    def generator():
        ''' Stream a zip file with all files using zipstream. '''
        def iterator():
            '''
            Replace the directory based iter of zipstream with an iter over all given
            files.
            '''
            # the actual contents
            for zipped_filename, file_id, open_io, file_size in files:
                if zipped_filename in streamed_files:
                    continue
                streamed_files.add(zipped_filename)

                # Write a file to the zipstream.
                try:
                    f = open_io(file_id)
                    try:
                        def iter_content():
                            while True:
                                data = f.read(1024 * 64)
                                if not data:
                                    break
                                yield data

                        yield dict(
                            arcname=zipped_filename, iterable=iter_content(),
                            buffer_size=file_size(file_id))
                    finally:
                        f.close()
                except KeyError:
                    # files that are not found, will not be returned
                    pass
                except Restricted:
                    # due to the streaming nature, we cannot raise 401 here
                    # we just leave it out in the download
                    pass

        compression = zipfile.ZIP_DEFLATED if compress else zipfile.ZIP_STORED
        zip_stream = zipstream.ZipFile(mode='w', compression=compression, allowZip64=True)
        zip_stream.paths_to_write = iterator()

        for chunk in zip_stream:
            yield chunk

    response = Response(stream_with_context(generator()), mimetype='application/zip')
    response.headers['Content-Disposition'] = 'attachment; filename={}'.format(zipfile_name)
    return response


def _query_api_url(*args, query: Dict[str, Any] = None):
    '''
    Creates a API URL.
    Arguments:
        *args: URL path segments after the API base URL
        query_string: A dict with query string parameters
    '''
    url = os.path.join(config.api_url(False), *args)
    if query is not None and len(query) > 0:
        url = '%s?%s' % (url, urlencode(query, doseq=True))

    return url


def query_api_python(query):
    '''
    Creates a string of python code to execute a search query to the repository using
    the requests library.
    '''
    query = _filter_api_query(query)
    url = _query_api_url('archive', 'query')
    return '''import requests
response = requests.post('{}', json={{
    'query': {{
{}
    }}
}})
data = response.json()'''.format(
        url,
        ',\n'.join([
            '        \'%s\': %s' % (key, pprint.pformat(value, compact=True))
            for key, value in query.items()])
    )


def _filter_api_query(query):
    def normalize_value(key, value):
        quantity = search.search_quantities.get(key)
        if quantity.many and not isinstance(value, list):
            return [value]
        elif isinstance(value, list) and len(value) == 1:
            return value[0]

        return value

    result = {
        key: normalize_value(key, value) for key, value in query.items()
        if key in search.search_quantities and (key != 'domain' or value != config.meta.default_domain)
    }

    for key in ['dft.optimade']:
        if key in query:
            result[key] = query[key]

    return result


def query_api_repo_url(query):
    '''
    Creates an encoded URL string access a search query on the repo api.
    '''
    query = dict(query)
    for to_delete in ['per_page', 'page', 'exclude']:
        if to_delete in query:
            del(query[to_delete])
    for key, value in dict(order_by=['upload_time'], order=['-1'], domain=['dft'], owner=['public']).items():
        if key in query and query[key] == value:
            del(query[key])
    return _query_api_url('repo', query=query)


def query_api_clientlib(**kwargs):
    '''
    Creates a string of python code to execute a search query on the archive using
    the client library.
    '''
    query = _filter_api_query(kwargs)

    out = io.StringIO()
    out.write('from nomad import client, config\n')
    out.write('config.client.url = \'%s\'\n' % config.api_url(ssl=False))
    out.write('results = client.query_archive(query={%s' % ('' if len(kwargs) == 0 else '\n'))
    out.write(',\n'.join([
        '    \'%s\': %s' % (key, pprint.pformat(value, compact=True))
        for key, value in query.items()]))
    out.write('})\n')
    out.write('print(results)\n')

    return out.getvalue()


def query_api_curl(query):
    '''
    Creates a string of curl command to execute a search query and download the respective
    archives in a .zip file.
    '''
    url = _query_api_url('archive', 'download', query=query)
    return 'curl "%s" --output nomad.zip' % url


def enable_gzip(level: int = 1, min_size: int = 1024):
    """
    Args:
        level: The gzip compression level from 1-9
        min_size: The minimum response size in bytes for which the compression
            will be enabled.
    """
    def inner(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            response = make_response(function(*args, **kwargs))
            if response.status_code == 200:
                accept_encoding = request.headers["Accept-Encoding"]
                content_length = int(response.headers["Content-Length"])
                if "gzip" in accept_encoding and content_length >= min_size:
                    data = response.data
                    data = gzip.compress(data, level)
                    response.data = data
                    response.headers['Content-Length'] = len(data)
                    response.headers["Content-Encoding"] = "gzip"
                    return response
            return response
        return wrapper
    return inner



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='cp2k_general_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='cp2k.general.nomadmetainfo.json'))


class x_cp2k_section_restart_information(MSection):
    '''
    Contains restart information for this calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_restart_information'))

    x_cp2k_restart_file_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the restart file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_restart_file_name'))

    x_cp2k_restarted_quantity_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of a restarted quantity.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_restarted_quantity_name'))


class x_cp2k_section_dbcsr(MSection):
    '''
    The DBCSR (Distributed Block Compressed Sparse Row) information.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_dbcsr'))

    x_cp2k_dbcsr_multiplication_driver = Quantity(
        type=str,
        shape=[],
        description='''
        DBCSR Multiplication driver
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_multiplication_driver'))

    x_cp2k_dbcsr_multrec_recursion_limit = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        DBCSR Multrec recursion limit
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_multrec_recursion_limit'))

    x_cp2k_dbcsr_multiplication_stack_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        DBCSR Multiplication stack size.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_multiplication_stack_size'))

    x_cp2k_dbcsr_multiplication_size_stacks = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        DBCSR Multiplication size of stacks.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_multiplication_size_stacks'))

    x_cp2k_dbcsr_use_subcommunicators = Quantity(
        type=str,
        shape=[],
        description='''
        Boolean indicating if subcommunicators are used.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_use_subcommunicators'))

    x_cp2k_dbcsr_use_mpi_combined_types = Quantity(
        type=str,
        shape=[],
        description='''
        Boolean indicating if MPI combined types are used.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_use_mpi_combined_types'))

    x_cp2k_dbcsr_use_mpi_memory_allocation = Quantity(
        type=str,
        shape=[],
        description='''
        Boolean indicating if MPI memory allocation is used.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_use_mpi_memory_allocation'))

    x_cp2k_dbcsr_use_communication_thread = Quantity(
        type=str,
        shape=[],
        description='''
        Boolean indicating if communication thread is used.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_use_communication_thread'))

    x_cp2k_dbcsr_communication_thread_load = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Load of the communication thread.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_dbcsr_communication_thread_load'))


class x_cp2k_section_global_settings(MSection):
    '''
    Global settings for this calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_global_settings'))

    x_cp2k_basis_set_filename = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the basis set file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_basis_set_filename'))

    x_cp2k_coordinate_filename = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the coordinate file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_coordinate_filename'))

    x_cp2k_geminal_filename = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the geminal file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_geminal_filename'))

    x_cp2k_mm_potential_filename = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the MM potential file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_mm_potential_filename'))

    x_cp2k_potential_filename = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the potential file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_potential_filename'))

    x_cp2k_method_name = Quantity(
        type=str,
        shape=[],
        description='''
        The method name for this run.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_method_name'))

    x_cp2k_preferred_fft_library = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the preferred FFT library.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_preferred_fft_library'))

    x_cp2k_preferred_diagonalization_library = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the preferred diagonalization library.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_preferred_diagonalization_library'))

    x_cp2k_run_type = Quantity(
        type=str,
        shape=[],
        description='''
        The run type for this calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_run_type'))


class x_cp2k_section_geometry_optimization_energy_reevaluation(MSection):
    '''
    Information for the energy re-evaluation at the end of an optimization procedure.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_geometry_optimization_energy_reevaluation'))


class x_cp2k_section_geometry_optimization_step(MSection):
    '''
    Contains information about the geometry optimization process for every optimization
    step.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_geometry_optimization_step'))

    x_cp2k_optimization_energy_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy change for this optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_energy_change'))

    x_cp2k_optimization_energy_decrease = Quantity(
        type=str,
        shape=[],
        description='''
        Whether there has been energy decrease. YES or NO.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_energy_decrease'))

    x_cp2k_optimization_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy for this optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_energy'))

    x_cp2k_optimization_gradient_convergence_limit = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Convergence criterium for the maximum force component of the current
        configuration.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_gradient_convergence_limit'))

    x_cp2k_optimization_max_gradient_convergence = Quantity(
        type=str,
        shape=[],
        description='''
        Whether there is convergence in max gradient. YES or NO.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_max_gradient_convergence'))

    x_cp2k_optimization_max_gradient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Max gradient for this optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_max_gradient'))

    x_cp2k_optimization_max_step_size = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum step size for this optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_max_step_size'))

    x_cp2k_optimization_method = Quantity(
        type=str,
        shape=[],
        description='''
        Optimization method for this step
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_method'))

    x_cp2k_optimization_rms_gradient_convergence = Quantity(
        type=str,
        shape=[],
        description='''
        Whether there is convergence in rms gradient. YES or NO.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_rms_gradient_convergence'))

    x_cp2k_optimization_rms_gradient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        RMS gradient for this optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_rms_gradient'))

    x_cp2k_optimization_rms_step_size_convergence = Quantity(
        type=str,
        shape=[],
        description='''
        Whether there is convergence in rms step size. YES or NO.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_rms_step_size_convergence'))

    x_cp2k_optimization_rms_step_size = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        RMS step size for this optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_rms_step_size'))

    x_cp2k_optimization_step_size_convergence_limit = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Convergence criterium for the maximum geometry change between the current and the
        last optimizer iteration.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_step_size_convergence_limit'))

    x_cp2k_optimization_step_size_convergence = Quantity(
        type=str,
        shape=[],
        description='''
        Whether there is convergence in step size. YES or NO.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_step_size_convergence'))

    x_cp2k_optimization_used_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Time used for this optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_optimization_used_time'))


class x_cp2k_section_geometry_optimization(MSection):
    '''
    CP2K geometry optimization information.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_geometry_optimization'))

    x_cp2k_section_geometry_optimization_energy_reevaluation = SubSection(
        sub_section=SectionProxy('x_cp2k_section_geometry_optimization_energy_reevaluation'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_geometry_optimization_energy_reevaluation'))

    x_cp2k_section_geometry_optimization_step = SubSection(
        sub_section=SectionProxy('x_cp2k_section_geometry_optimization_step'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_geometry_optimization_step'))


class x_cp2k_section_maximum_angular_momentum(MSection):
    '''
    Contains the maximum angular momentum values used in the calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_maximum_angular_momentum'))

    x_cp2k_local_part_of_gth_pseudopotential = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maximum angular momentum of the local part of the GTH pseudopotential.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_local_part_of_gth_pseudopotential'))

    x_cp2k_non_local_part_of_gth_pseudopotential = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maximum angular momentum of the non-local part of the GTH pseudopotential.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_non_local_part_of_gth_pseudopotential'))

    x_cp2k_orbital_basis_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maximum angular momentum of orbital basis functions.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_orbital_basis_functions'))


class x_cp2k_section_program_information(MSection):
    '''
    Contains information about the software version used for this run.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_program_information'))

    x_cp2k_input_filename = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the CP2K input file that produced this calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_input_filename'))

    x_cp2k_program_compilation_datetime = Quantity(
        type=str,
        shape=[],
        description='''
        The time when this program was compiled.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_program_compilation_datetime'))

    x_cp2k_svn_revision = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The SVN revision number.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_svn_revision'))


class x_cp2k_section_quickstep_calculation(MSection):
    '''
    Section for a CP2K QuickStep calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_quickstep_calculation'))

    x_cp2k_atom_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='newton',
        description='''
        Forces acting on the atoms in this Quickstep calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_atom_forces'))

    x_cp2k_electronic_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Self-consistent electronic kinetic energy calculated with Quickstep
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_electronic_kinetic_energy'))

    x_cp2k_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the total energy (nuclei + electrons) calculated with Quickstep.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_energy_total'))

    x_cp2k_quickstep_converged = Quantity(
        type=bool,
        shape=[],
        description='''
        Boolean indicating whether the quickstep SCF cycle converged or not.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_quickstep_converged'))

    x_cp2k_section_scf_iteration = SubSection(
        sub_section=SectionProxy('x_cp2k_section_scf_iteration'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_scf_iteration'))

    x_cp2k_section_stress_tensor = SubSection(
        sub_section=SectionProxy('x_cp2k_section_stress_tensor'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_stress_tensor'))


class x_cp2k_section_scf_iteration(MSection):
    '''
    Section for a CP2K QuickStep SCF iteration.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_scf_iteration'))

    x_cp2k_energy_change_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        At each self-consistent field (SCF) iteration, change of total energy with respect
        to the previous SCF iteration.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_energy_change_scf_iteration'))

    x_cp2k_energy_total_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Total electronic energy calculated with XC_method during the self-consistent field
        (SCF) iterations.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_energy_total_scf_iteration'))

    x_cp2k_energy_XC_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Exchange-correlation (XC) energy during the self-consistent field (SCF) iteration.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_energy_XC_scf_iteration'))


class x_cp2k_section_startinformation(MSection):
    '''
    Contains information about the starting conditions for this run.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_startinformation'))

    x_cp2k_start_time = Quantity(
        type=str,
        shape=[],
        description='''
        The starting time for this run.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_start_time'))

    x_cp2k_start_host = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the host machine this calculation started on.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_start_host'))

    x_cp2k_start_user = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the user at the start of the calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_start_user'))

    x_cp2k_start_id = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The process id at the start of this run.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_start_id'))

    x_cp2k_start_path = Quantity(
        type=str,
        shape=[],
        description='''
        The path where this calculation started.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_start_path'))


class x_cp2k_section_end_information(MSection):
    '''
    Contains information about the ending conditions for this run.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_end_information'))

    x_cp2k_end_time = Quantity(
        type=str,
        shape=[],
        description='''
        The ending time for this run.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_end_time'))

    x_cp2k_end_host = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the host machine this calculation ended on.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_end_host'))

    x_cp2k_end_user = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the user at the end of the calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_end_user'))

    x_cp2k_end_id = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The process id at the end of this run.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_end_id'))

    x_cp2k_end_path = Quantity(
        type=str,
        shape=[],
        description='''
        The path where this calculation ended.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_end_path'))


class x_cp2k_section_stress_tensor(MSection):
    '''
    Section for stress tensor information.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_stress_tensor'))

    x_cp2k_stress_tensor_determinant = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The determinant of the stress tensor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_stress_tensor_determinant'))

    x_cp2k_stress_tensor_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        The eigenvalues of the stress tensor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_stress_tensor_eigenvalues'))

    x_cp2k_stress_tensor_eigenvectors = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        The eigenvectors of the stress tensor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_stress_tensor_eigenvectors'))

    x_cp2k_stress_tensor_one_third_of_trace = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        1/3 of the trace of the stress tensor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_stress_tensor_one_third_of_trace'))

    x_cp2k_stress_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='pascal',
        description='''
        A final value of the stress tensor in a Quickstep calculation
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_stress_tensor'))


class x_cp2k_section_total_numbers(MSection):
    '''
    The total number of different entities in the calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_total_numbers'))

    x_cp2k_atomic_kinds = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The number of atomic kinds in the calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_atomic_kinds'))

    x_cp2k_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The number of atoms in the calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_atoms'))

    x_cp2k_cartesian_basis_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The number of Cartesian basis functions.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_cartesian_basis_functions'))

    x_cp2k_primitive_cartesian_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The number of primitive Cartesian functions.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_primitive_cartesian_functions'))

    x_cp2k_shell_sets = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The number of shell sets in the calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_shell_sets'))

    x_cp2k_shells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The number of shells.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_shells'))

    x_cp2k_spherical_basis_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The number of Spherical basis functions.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_spherical_basis_functions'))


class x_cp2k_section_md(MSection):
    '''
    CP2K Molecular Dynamics information.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_md'))

    x_cp2k_section_md_step = SubSection(
        sub_section=SectionProxy('x_cp2k_section_md_step'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_md_step'))


class x_cp2k_section_md_settings(MSection):
    '''
    Settings for CP2K Molecular Dynamics.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_md_settings'))

    x_cp2k_md_ensemble_type = Quantity(
        type=str,
        shape=[],
        description='''
        The ensemble type in molecular dynamics.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_ensemble_type'))

    x_cp2k_md_time_step = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Time step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_time_step'))

    x_cp2k_md_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Thermostat target temperature.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_target_temperature'))

    x_cp2k_md_target_temperature_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Target temperature tolerance.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_target_temperature_tolerance'))

    x_cp2k_md_print_frequency = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The print frequency of molecular dynamics information in the CP2K output file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_print_frequency'))

    x_cp2k_md_coordinates_print_frequency = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Print frequency for the coordinate file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_coordinates_print_frequency'))

    x_cp2k_md_coordinates_filename = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the coordinate file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_coordinates_filename'))

    x_cp2k_md_velocities_print_frequency = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Print frequency for the velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_velocities_print_frequency'))

    x_cp2k_md_velocities_filename = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_velocities_filename'))

    x_cp2k_md_energies_print_frequency = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Print frequency for the energies file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_energies_print_frequency'))

    x_cp2k_md_energies_filename = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the energies file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_energies_filename'))

    x_cp2k_md_dump_print_frequency = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Print frequency for the dump file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_dump_print_frequency'))

    x_cp2k_md_dump_filename = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the dump file.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_dump_filename'))

    x_cp2k_md_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Target pressure for the barostat.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_target_pressure'))

    x_cp2k_md_barostat_time_constant = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Barostat time constant.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_barostat_time_constant'))

    x_cp2k_md_simulation_cell_print_frequency = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Simulation cell print frequency.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_simulation_cell_print_frequency'))

    x_cp2k_md_simulation_cell_filename = Quantity(
        type=str,
        shape=[],
        description='''
        Simulation cell filename.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_simulation_cell_filename'))

    x_cp2k_md_number_of_time_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of requested time steps in molecular dynamics.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_number_of_time_steps'))


class x_cp2k_section_md_step(MSection):
    '''
    Information from MD step.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_md_step'))

    x_cp2k_md_potential_energy_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous potential energy in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_potential_energy_instantaneous'))

    x_cp2k_md_potential_energy_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average potential energy for an MD step. Averaged over this and the previous
        steps.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_potential_energy_average'))

    x_cp2k_md_kinetic_energy_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous kinetic energy in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_kinetic_energy_instantaneous'))

    x_cp2k_md_kinetic_energy_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average kinetic energy in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_kinetic_energy_average'))

    x_cp2k_md_energy_drift_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous energy drift in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_energy_drift_instantaneous'))

    x_cp2k_md_energy_drift_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average kinetic energy in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_energy_drift_average'))

    x_cp2k_md_temperature_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous temperature in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_temperature_instantaneous'))

    x_cp2k_md_temperature_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average temperature in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_temperature_average'))

    x_cp2k_md_barostat_temperature_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous barostat temperature in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_barostat_temperature_instantaneous'))

    x_cp2k_md_barostat_temperature_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average barostat temperature in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_barostat_temperature_average'))

    x_cp2k_md_pressure_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous pressure in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_pressure_instantaneous'))

    x_cp2k_md_pressure_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average pressure in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_pressure_average'))

    x_cp2k_md_cpu_time_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous CPU time for this step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cpu_time_instantaneous'))

    x_cp2k_md_cpu_time_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average CPU time for this step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cpu_time_average'))

    x_cp2k_md_step_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Step number.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_step_number'))

    x_cp2k_md_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation time for this step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_time'))

    x_cp2k_md_conserved_quantity = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Conserved quantity for this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_conserved_quantity'))

    x_cp2k_md_volume_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous cell volume in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_volume_instantaneous'))

    x_cp2k_md_volume_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average cell volume in an MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_volume_average'))

    x_cp2k_md_cell_length_a_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous cell vector a length in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_length_a_instantaneous'))

    x_cp2k_md_cell_length_a_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average cell vector a length in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_length_a_average'))

    x_cp2k_md_cell_length_b_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous cell vector b length in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_length_b_instantaneous'))

    x_cp2k_md_cell_length_b_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average cell vector b length in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_length_b_average'))

    x_cp2k_md_cell_length_c_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous cell vector c length in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_length_c_instantaneous'))

    x_cp2k_md_cell_length_c_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average cell vector c length in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_length_c_average'))

    x_cp2k_md_cell_angle_a_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous cell vector a angle in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_angle_a_instantaneous'))

    x_cp2k_md_cell_angle_a_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average cell vector a angle in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_angle_a_average'))

    x_cp2k_md_cell_angle_b_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous cell vector b angle in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_angle_b_instantaneous'))

    x_cp2k_md_cell_angle_b_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average cell vector b angle in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_angle_b_average'))

    x_cp2k_md_cell_angle_c_instantaneous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Instantaneous cell vector c angle in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_angle_c_instantaneous'))

    x_cp2k_md_cell_angle_c_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average cell vector c angle in this MD step.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_md_cell_angle_c_average'))


class x_cp2k_section_quickstep_settings(MSection):
    '''
    Quickstep settings.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_quickstep_settings'))

    x_cp2k_planewave_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The plane-wave cutoff for the auxiliary basis.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_planewave_cutoff'))

    x_cp2k_spin_restriction = Quantity(
        type=str,
        shape=[],
        description='''
        Indicates the restriction applied for the spin (e.g. RKS).
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_spin_restriction'))

    x_cp2k_quickstep_method = Quantity(
        type=str,
        shape=[],
        description='''
        The method used for the Quickstep calculations (GPW, GAPW).
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_quickstep_method'))

    x_cp2k_section_maximum_angular_momentum = SubSection(
        sub_section=SectionProxy('x_cp2k_section_maximum_angular_momentum'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_maximum_angular_momentum'))

    x_cp2k_section_total_numbers = SubSection(
        sub_section=SectionProxy('x_cp2k_section_total_numbers'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_total_numbers'))

    x_cp2k_section_atomic_kinds = SubSection(
        sub_section=SectionProxy('x_cp2k_section_atomic_kinds'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_atomic_kinds'))


class x_cp2k_section_atomic_kinds(MSection):
    '''
    Information about all the atomic kinds in this Quickstep calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_atomic_kinds'))

    x_cp2k_section_atomic_kind = SubSection(
        sub_section=SectionProxy('x_cp2k_section_atomic_kind'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_atomic_kind'))


class x_cp2k_section_vdw_settings(MSection):
    '''
    Van der Waals settings.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_settings'))

    x_cp2k_vdw_type = Quantity(
        type=str,
        shape=[],
        description='''
        Type of the van der Waals method.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_type'))

    x_cp2k_vdw_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the van der Waals method.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_name'))

    x_cp2k_vdw_bj_damping_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the BJ damping method.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_bj_damping_name'))

    x_cp2k_vdw_cutoff_radius = Quantity(
        type=str,
        shape=[],
        description='''
        Cutoff radius of the van der Waals method.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_cutoff_radius'))

    x_cp2k_section_vdw_d2_settings = SubSection(
        sub_section=SectionProxy('x_cp2k_section_vdw_d2_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_d2_settings'))

    x_cp2k_section_vdw_d3_settings = SubSection(
        sub_section=SectionProxy('x_cp2k_section_vdw_d3_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_d3_settings'))


class x_cp2k_section_vdw_d2_settings(MSection):
    '''
    D2 settings.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_d2_settings'))

    x_cp2k_vdw_scaling_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Scaling factor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_scaling_factor'))

    x_cp2k_vdw_damping_factor = Quantity(
        type=str,
        shape=[],
        description='''
        Exponential damping prefactor for the van der Waals method.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_damping_factor'))

    x_cp2k_section_vdw_element_settings = SubSection(
        sub_section=SectionProxy('x_cp2k_section_vdw_element_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_element_settings'))


class x_cp2k_section_vdw_element_settings(MSection):
    '''
    Contains element-specific Van der Waals settings.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_element_settings'))

    x_cp2k_vdw_parameter_element_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the element.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_parameter_element_name'))

    x_cp2k_vdw_parameter_c6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        C6 parameter.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_parameter_c6'))

    x_cp2k_vdw_parameter_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Radius parameter.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_parameter_radius'))


class x_cp2k_section_vdw_d3_settings(MSection):
    '''
    D3 settings.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_d3_settings'))

    x_cp2k_vdw_s6_scaling_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        S6 scaling factor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_s6_scaling_factor'))

    x_cp2k_vdw_sr6_scaling_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SR6 scaling factor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_sr6_scaling_factor'))

    x_cp2k_vdw_s8_scaling_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        S8 scaling factor.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_s8_scaling_factor'))

    x_cp2k_vdw_cn_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Cutoff for CN calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_vdw_cn_cutoff'))


class x_cp2k_section_atomic_kind(MSection):
    '''
    Information one atomic kind.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_atomic_kind'))

    x_cp2k_kind_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The atomic kind number. For each element there may be multiple kinds specified.
        This number differentiates them. Not the atomic number.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_kind_number'))

    x_cp2k_kind_label = Quantity(
        type=str,
        shape=[],
        description='''
        The label for this atomic kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_kind_label'))

    x_cp2k_kind_number_of_atoms = Quantity(
        type=str,
        shape=[],
        description='''
        The number of atoms with this kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_kind_number_of_atoms'))

    x_cp2k_section_kind_basis_set = SubSection(
        sub_section=SectionProxy('x_cp2k_section_kind_basis_set'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_kind_basis_set'))


class x_cp2k_section_kind_basis_set(MSection):
    '''
    Description of the basis set used for this kind.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cp2k_section_kind_basis_set'))

    x_cp2k_kind_basis_set_name = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the orbital basis set used for this kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_kind_basis_set_name'))

    x_cp2k_basis_set_number_of_orbital_shell_sets = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of orbital shell sets.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_basis_set_number_of_orbital_shell_sets'))

    x_cp2k_basis_set_number_of_orbital_shells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of orbital shells.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_basis_set_number_of_orbital_shells'))

    x_cp2k_basis_set_number_of_primitive_cartesian_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of primitive Cartesian functions.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_basis_set_number_of_primitive_cartesian_functions'))

    x_cp2k_basis_set_number_of_cartesian_basis_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of Cartesian basis functions.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_basis_set_number_of_cartesian_basis_functions'))

    x_cp2k_basis_set_number_of_spherical_basis_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of spherical basis functions.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_basis_set_number_of_spherical_basis_functions'))

    x_cp2k_basis_set_norm_type = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Norm type.
        ''',
        a_legacy=LegacyDefinition(name='x_cp2k_basis_set_norm_type'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_cp2k_section_restart_information = SubSection(
        sub_section=SectionProxy('x_cp2k_section_restart_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_restart_information'))

    x_cp2k_section_dbcsr = SubSection(
        sub_section=SectionProxy('x_cp2k_section_dbcsr'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_dbcsr'))

    x_cp2k_section_global_settings = SubSection(
        sub_section=SectionProxy('x_cp2k_section_global_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_global_settings'))

    x_cp2k_section_program_information = SubSection(
        sub_section=SectionProxy('x_cp2k_section_program_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_program_information'))

    x_cp2k_section_quickstep_calculation = SubSection(
        sub_section=SectionProxy('x_cp2k_section_quickstep_calculation'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_quickstep_calculation'))

    x_cp2k_section_startinformation = SubSection(
        sub_section=SectionProxy('x_cp2k_section_startinformation'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_startinformation'))

    x_cp2k_section_end_information = SubSection(
        sub_section=SectionProxy('x_cp2k_section_end_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_end_information'))


class section_frame_sequence(public.section_frame_sequence):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_frame_sequence'))

    x_cp2k_section_geometry_optimization = SubSection(
        sub_section=SectionProxy('x_cp2k_section_geometry_optimization'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_geometry_optimization'))

    x_cp2k_section_md = SubSection(
        sub_section=SectionProxy('x_cp2k_section_md'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_md'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_cp2k_section_md_settings = SubSection(
        sub_section=SectionProxy('x_cp2k_section_md_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_md_settings'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_cp2k_section_quickstep_settings = SubSection(
        sub_section=SectionProxy('x_cp2k_section_quickstep_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_quickstep_settings'))

    x_cp2k_section_vdw_settings = SubSection(
        sub_section=SectionProxy('x_cp2k_section_vdw_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cp2k_section_vdw_settings'))


m_package.__init_metainfo__()



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
import re
p = re.compile(
      '((?P<x_name>(GGA|LDA|MGGA|HF|HYB_MGGA)_X.*)|(?P<c_name>(GGA|LDA|MGGA)_C.*))')

short_names = {
        'LDA.RPA': 'LDA_X+LDA_C_RPA',
        'LDA.PZ': 'LDA_X+LDA_C_PZ',
        'LDA.PW': 'LDA_X+LDA_C_PW',
        'GGA.PW91': 'GGA_X_PW91+GGA_C_PW91',
        'GGA.PBE': 'GGA_X_PBE+GGA_C_PBE',
        'GGA.PBES': 'GGA_X_PBE_SOL+GGA_C_PBE_SOL',
        'GGA.RPBE': 'GGA_X_RPBE+GGA_C_PBE',
        'BLYP': 'GGA_X_B88+GGA_C_LYP',
        'HCTH407': 'GGA_XC_HCTH_407',
        'WC': 'GGA_X_WC+GGA_C_PBE',
        'AM05': 'GGA_X_AM05+GGA_C_AM05',
        'mBEEF': 'MGGA_X_MBEEF+GGA_C_PBE_SOL'}


def get_libxc_name(name):
    if name in short_names:
        libxc_name = short_names[name]
    else:
        libxc_name = name
    return libxc_name


def get_libxc_xc_names(name):
    """get dictionary with
       x_name: Exchange name (None if xc_name is not None)
       c_name: Correlation name (-||-)
       xc_name: XC name
    """
    name = get_libxc_name(name)
    xc = {'xc_name': None,
          'x_name' : None,
          'c_name': None}

    if '_XC_' in name:
        xc['xc_name'] = name
        return xc

    if '+' in name:
        s = name.split('+')
        xc['x_name'] = s[0]
        xc['c_name'] = s[1]
        return xc

    m = re.search(p, name)
    if m is not None: # it is either a correlation or exchange functional
        xc.update(m.groupdict())
        return xc

    xc['xc_name'] = name  # for something like BEEF-vdW
    return xc


if __name__ == '__main__':
    print(get_libxc_xc_names('LDA.PZ'))
    print(get_libxc_xc_names('LDA_X'))
    print(get_libxc_xc_names('GGA_X_PBE'))



import math

import pytest

from .utils import get_min_ver

min_ver = get_min_ver("numpy")
numpy = pytest.importorskip(
    "numpy",
    minversion=min_ver,
    reason=f"numpy must be installed with minimum version {min_ver} for these tests to"
    " be able to run",
)

from optimade.adapters.structures.utils import (
    fractional_coordinates,
    pad_cell,
    pad_positions,
    scaled_cell,
)


# TODO: Add tests for cell_to_cellpar, unit_vector, cellpar_to_cell


def test_pad_positions(null_position_structure):
    """Make sure None values in cartesian_site_positions are converted to padding float value"""
    positions, padded_position = pad_positions(
        null_position_structure.attributes.cartesian_site_positions
    )

    assert not any(value is None for vector in positions for value in vector)
    assert padded_position

    positions, padded_position = pad_positions(positions)

    assert not any(value is None for vector in positions for value in vector)
    assert not padded_position


def test_pad_cell(null_lattice_vector_structure):
    """Make sure None values in lattice_vectors are converted to padding float value"""
    lattice_vectors, padded_cell = pad_cell(
        null_lattice_vector_structure.attributes.lattice_vectors
    )

    assert not any(value is None for vector in lattice_vectors for value in vector)
    assert padded_cell

    lattice_vectors, padded_cell = pad_cell(lattice_vectors)

    assert not any(value is None for vector in lattice_vectors for value in vector)
    assert not padded_cell


def test__pad_iter_of_iters():
    """Test _pad_iter_of_iters"""
    iterable = [(0.0,) * 3, (0.0,) * 3, (None,) * 3]

    padded_iterable, padded_iterable_bool = pad_cell(iterable)

    assert padded_iterable_bool
    assert all(math.isnan(_) for _ in padded_iterable[-1])
    for i in range(2):
        assert padded_iterable[i] == (0.0,) * 3

    for valid_padding_value in (3.0, 3, "3", "3.0"):
        padded_iterable, padded_iterable_bool = pad_cell(iterable, valid_padding_value)

        assert padded_iterable_bool
        assert padded_iterable[-1] == (float(valid_padding_value),) * 3
        for i in range(2):
            assert padded_iterable[i] == (0.0,) * 3

    # Since nan != nan, the above for-loop cannot be used for nan
    valid_padding_value = "nan"
    padded_iterable, padded_iterable_bool = pad_cell(iterable, valid_padding_value)

    assert padded_iterable_bool
    assert all(math.isnan(_) for _ in padded_iterable[-1])
    assert all(math.isnan(_) for _ in (float(valid_padding_value),) * 3)
    for i in range(2):
        assert padded_iterable[i] == (0.0,) * 3

    invalid_padding_value = "x"
    padded_iterable, padded_iterable_bool = pad_cell(iterable, invalid_padding_value)

    assert padded_iterable_bool
    assert all(math.isnan(_) for _ in padded_iterable[-1])
    for i in range(2):
        assert padded_iterable[i] == (0.0,) * 3


def test_scaled_cell_and_fractional_coordinates(structures):
    """Make sure these two different calculations arrive at the same result"""
    for structure in structures:
        scale = scaled_cell(structure.lattice_vectors)
        scale = numpy.asarray(scale)
        cartesian_positions = numpy.asarray(structure.cartesian_site_positions)
        scaled_fractional_positions = (scale.T @ cartesian_positions.T).T
        for i in range(3):
            scaled_fractional_positions[:, i] %= 1.0
            scaled_fractional_positions[:, i] %= 1.0
        scaled_fractional_positions = [
            tuple(position) for position in scaled_fractional_positions
        ]

        calculated_fractional_positions = fractional_coordinates(
            cell=structure.lattice_vectors,
            cartesian_positions=structure.cartesian_site_positions,
        )

        for scaled_position, calculated_position in zip(
            scaled_fractional_positions, calculated_fractional_positions
        ):
            assert scaled_position == pytest.approx(calculated_position)


def test_scaled_cell_consistency(structure):
    """Test scaled_cell's PDB-designated validation: inverse of det(SCALE) = Volume of cell"""
    # Manual calculation of volume = |a_1 . (a_2 x a_3)|
    a_1 = structure.lattice_vectors[0]
    a_2 = structure.lattice_vectors[1]
    a_3 = structure.lattice_vectors[2]
    a_mid_0 = a_2[1] * a_3[2] - a_2[2] * a_3[1]
    a_mid_1 = a_2[2] * a_3[0] - a_2[0] * a_3[2]
    a_mid_2 = a_2[0] * a_3[1] - a_2[1] * a_3[0]
    volume_from_cellpar = abs(a_1[0] * a_mid_0 + a_1[1] * a_mid_1 + a_1[2] * a_mid_2)

    scale = scaled_cell(structure.lattice_vectors)
    volume_from_scale = 1 / numpy.linalg.det(scale)

    assert volume_from_scale == pytest.approx(volume_from_cellpar)



""" This module contains the ImplementationValidator class and corresponding command line tools. """
# pylint: disable=import-outside-toplevel
from .validator import ImplementationValidator

__all__ = ["ImplementationValidator", "validate"]


def validate():
    import argparse
    import sys
    import os
    import traceback

    parser = argparse.ArgumentParser(
        prog="optimade_validator",
        description="""Tests OPTIMADE implementations for compliance with the optimade-python-tools models.

    - To test an entire implementation (at say example.com/optimade/v1) for all required/available endpoints:

        $ optimade_validator http://example.com/optimade/v1

    - To test a particular response of an implementation against a particular model:

        $ optimade_validator http://example.com/optimade/v1/structures/id=1234 --as_type structure

    - To test a particular response of an implementation against a particular model:

        $ optimade_validator http://example.com/optimade/v1/structures --as_type structures
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "base_url",
        nargs="?",
        default="http://localhost:5000/v0",
        help=(
            "The base URL of the OPTIMADE implementation to point at, "
            "e.g. 'http://example.com/optimade/v1' or 'http://localhost:5000/v1"
        ),
    )
    parser.add_argument(
        "--verbosity", "-v", type=int, default=0, help="The verbosity of the output"
    )
    parser.add_argument(
        "--as_type",
        "-a",
        type=str,
        help=(
            "Validate the request URL with the provided type, rather than scanning the entire implementation e.g. "
            "optimade_validator `http://example.com/optimade/v1/structures/0 --as_type structures`"
        ),
    )
    parser.add_argument(
        "--index",
        action="store_true",
        help=(
            "Flag for whether the specified OPTIMADE implementation is an Index meta-database or not. [default=False]"
        ),
    )

    args = vars(parser.parse_args())

    if os.environ.get("OPTIMADE_VERBOSITY") is not None:
        try:
            args["verbosity"] = int(os.environ.get("OPTIMADE_VERBOSITY", 0))
        except (TypeError, ValueError):
            pass

    valid_types = [
        "info",
        "info/references",
        "info/structures",
        "links",
        "references",
        "reference",
        "structures",
        "structure",
    ]
    if args["as_type"] is not None and args["as_type"] not in valid_types:
        sys.exit(f"{args['as_type']} is not a valid type, must be one of {valid_types}")

    validator = ImplementationValidator(
        base_url=args["base_url"],
        verbosity=args["verbosity"],
        as_type=args["as_type"],
        index=args["index"],
    )

    try:
        validator.main()
    # catch and print internal exceptions, exiting with non-zero error code
    except Exception:
        traceback.print_exc()

    if validator.valid is None:
        sys.exit(2)
    elif not validator.valid:
        sys.exit(1)



""" This class contains patched versions of the OPTIMADE models as
a workaround for the response field workaround of allowing responses
to contain bare dictionaries. These allow the validator to print detailed
validation responses.

"""

from typing import List, Optional, Dict, Any

from pydantic import Field

from optimade.models.optimade_json import Success
from optimade.models import (
    ResponseMeta,
    EntryResource,
    LinksResource,
    ReferenceResource,
    StructureResource,
)


class ValidatorLinksResponse(Success):
    meta: ResponseMeta = Field(...)
    data: List[LinksResource] = Field(...)


class ValidatorEntryResponseOne(Success):
    meta: ResponseMeta = Field(...)
    data: EntryResource = Field(...)
    included: Optional[List[Dict[str, Any]]] = Field(...)


class ValidatorEntryResponseMany(Success):
    meta: ResponseMeta = Field(...)
    data: List[EntryResource] = Field(...)
    included: Optional[List[Dict[str, Any]]] = Field(...)


class ValidatorReferenceResponseOne(ValidatorEntryResponseOne):
    data: ReferenceResource = Field(...)


class ValidatorReferenceResponseMany(ValidatorEntryResponseMany):
    data: List[ReferenceResource] = Field(...)


class ValidatorStructureResponseOne(ValidatorEntryResponseOne):
    data: StructureResource = Field(...)


class ValidatorStructureResponseMany(ValidatorEntryResponseMany):
    data: List[StructureResource] = Field(...)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Context():
    """A simple class for holding the context related to an Encylopedia entry.
    """
    def __init__(
        self,
        material_type: str,
        method_type: str,
        calc_type: str,
        representative_system,
        representative_method,
        representative_scc,
        representative_scc_idx,
    ):
        self.material_type = material_type
        self.method_type = method_type
        self.calc_type = calc_type
        self.representative_system = representative_system
        self.representative_method = representative_method
        self.representative_scc = representative_scc
        self.representative_scc_idx = representative_scc_idx
        self.greatest_common_divisor: int = None



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import fractions
import itertools
from math import gcd as gcd
from functools import reduce
from typing import List, Dict, Tuple, Any, Union

import numpy as np
from scipy.spatial import Voronoi  # pylint: disable=no-name-in-module
from matid.symmetry import WyckoffSet

from nomad.aflow_prototypes import aflow_prototypes
from nomad.constants import atomic_masses


def get_summed_atomic_mass(atomic_numbers: np.ndarray) -> float:
    """Calculates the summed atomic mass for the given atomic numbers.

    Args:
        atomic_numbers: Array of valid atomic numbers

    Returns:
        The atomic mass in kilograms.
    """
    # It is assumed that the atomic numbers are valid at this point.
    mass = np.sum(atomic_masses[atomic_numbers])
    return mass


def get_volume(parallelepiped: np.ndarray) -> float:
    """Calculates a volume of the given parallelepiped.

    Args:
        cell: The parallellepiped as 3x3 matrix with cell basis vectors as
        rows.

    Returns:
        The cell volume.
    """
    return np.abs(np.linalg.det(parallelepiped))


def find_match(pos: np.array, positions: np.array, eps: float) -> Union[int, None]:
    """Attempts to find a position within a larger list of positions.

    Args:
        pos: The point to search for
        positions: The points within which the search is performed.
        eps: Match tolerance.

    Returns:
        Index of the matched position or None if match not found.
    """
    displacements = positions - pos
    distances = np.linalg.norm(displacements, axis=1)
    min_arg = np.argmin(distances)
    min_value = distances[min_arg]
    if min_value <= eps:
        return min_arg
    else:
        return None


def get_symmetry_string(space_group: int, wyckoff_sets: List[WyckoffSet], is_2d: bool = False) -> str:
    """Used to serialize symmetry information into a string. The Wyckoff
    positions are assumed to be normalized and ordered as is the case if using
    the matid-library.

    Args:
        space_group: 3D space group number
        wyckoff_sets: Wyckoff sets that map a Wyckoff letter to related
            information
        is_2d: Whether the symmetry information is analyzed from a 2D
            structure. If true, a prefix is added to the string to distinguish
            2D from 3D.

    Returns:
        A string that encodes the symmetry properties of an atomistic
        structure.
    """
    wyckoff_strings = []
    for group in wyckoff_sets:
        element = group.element
        wyckoff_letter = group.wyckoff_letter
        n_atoms = len(group.indices)
        i_string = "{} {} {}".format(element, wyckoff_letter, n_atoms)
        wyckoff_strings.append(i_string)
    wyckoff_string = ", ".join(sorted(wyckoff_strings))
    if is_2d:
        string = "2D {} {}".format(space_group, wyckoff_string)
    else:
        string = "{} {}".format(space_group, wyckoff_string)

    return string


def get_lattice_parameters(normalized_cell: np.ndarray) -> np.ndarray:
    """Calculate the lattice parameters for the normalized cell.

    Args:
        normalized_cell: The normalized cell as a 3x3 array. Each row is a
            basis vector.

    Returns:
        Six parameters a, b, c, alpha, beta, gamma (in this order) as a numpy
        array. Here is an explanation of each parameter:

        a = length of first basis vector
        b = length of second basis vector
        c = length of third basis vector
        alpha = angle between b and c
        beta  = angle between a and c
        gamma = angle between a and b
    """
    if normalized_cell is None:
        return None

    # Lengths
    lengths = np.linalg.norm(normalized_cell, axis=1)
    a, b, c = lengths

    # Angles
    angles = np.zeros(3)
    for i in range(3):
        j = (i + 1) % 3
        k = (i + 2) % 3
        angles[i] = np.dot(
            normalized_cell[j],
            normalized_cell[k]) / (lengths[j] * lengths[k])
    angles = np.clip(angles, -1.0, 1.0)
    alpha, beta, gamma = np.arccos(angles)

    return [a, b, c, alpha, beta, gamma]


def get_hill_decomposition(atom_labels: np.ndarray, reduced: bool = False) -> Tuple[List[str], List[int]]:
    """Given a list of atomic labels, returns the chemical formula using the
    Hill system (https://en.wikipedia.org/wiki/Hill_system) with an exception
    for binary ionic compounds where the cation is always given first.

    Args:
        atom_labels: Atom labels.
        reduced: Whether to divide the number of atoms by the greatest common
            divisor

    Returns:
        An ordered list of chemical symbols and the corresponding counts.
    """
    # Count occurancy of elements
    names = []
    counts = []
    unordered_names, unordered_counts = np.unique(atom_labels, return_counts=True)
    element_count_map = dict(zip(unordered_names, unordered_counts))

    # Apply basic Hill system:
    # 1. Is Carbon part of the system?
    if "C" in element_count_map:
        names.append("C")
        counts.append(element_count_map["C"])
        del element_count_map['C']

        # 1a. add hydrogren
        if "H" in element_count_map:
            names.append("H")
            counts.append(element_count_map["H"])
            del element_count_map["H"]

    # 2. all remaining elements in alphabetic order
    for element in sorted(element_count_map):
        names.append(element)
        counts.append(element_count_map[element])

    # 3. Binary ionic compounds: cation first, anion second
    # If any of the most electronegative elements is first
    # by alphabetic order, we move it to second
    if len(counts) == 2 and names != ["C", "H"]:
        order = {
            "F": 1,
            "O": 2,
            "N": 3,
            "Cl": 4,
            "Br": 5,
            "C": 6,
            "Se": 7,
            "S": 8,
            "I": 9,
            "As": 10,
            "H": 11,
            "P": 12,
            "Ge": 13,
            "Te": 14,
            "B": 15,
            "Sb": 16,
            "Po": 17,
            "Si": 18,
            "Bi": 19
        }
        if (names[0] in order):
            if (names[1] in order):
                if(order[names[0]] < order[names[1]]):
                    # For non-metals:
                    # Swap symbols and counts if first element
                    # is more electronegative than the second one,
                    # because the more electronegative element is the anion
                    names[0], names[1] = names[1], names[0]
                    counts[0], counts[1] = counts[1], counts[0]
            else:
                # Swap symbols and counts always if second element
                # is any other element,i.e.,
                # put non-metal last because it is the anion
                names[0], names[1] = names[1], names[0]
                counts[0], counts[1] = counts[1], counts[0]

    # TODO: implement all further exceptions regarding ordering
    #       in chemical formulas:
    #         - ionic compounds (ordering wrt to ionization)
    #         - oxides, acids, hydroxides...

    # Reduce if requested
    if reduced:
        greatest_common_divisor = reduce(gcd, counts)
        counts = np.array(counts) / greatest_common_divisor

    return names, counts


def get_formula_string(symbols: List[str], counts: List[int]) -> str:
    """Used to form a single formula string from a list of chemical species and
    their counts.

    Args:
        symbols: List of chemical species
        counts: List of chemical species occurences

    Returns:
        The formula as a string.
    """
    formula = ""
    for symbol, count in zip(symbols, counts):
        if count > 1:
            formula += "%s%d" % (symbol, count)
        else:
            formula += symbol
    return formula


def get_normalized_wyckoff(atomic_numbers: np.array, wyckoff_letters: np.array) -> Dict[str, Dict[str, int]]:
    """Returns a normalized Wyckoff sequence for the given atomic numbers and
    corresponding wyckoff letters. In a normalized sequence the chemical
    species are "anonymized" by replacing them with upper case alphabets.

    Args:
        atomic_numbers: Array of atomic numbers.
        wyckoff_letters: Array of Wyckoff letters as strings.

    Returns:
        Returns a dictionary that maps each present Wyckoff letter to a
        dictionary. The dictionary contains the number of atoms for each
        species, where the species names have been anomymized in the form
        "X_<index>".
    """
    # Count the occurrence of each chemical species
    atom_count: Dict[int, int] = {}
    for atomic_number in atomic_numbers:
        atom_count[atomic_number] = atom_count.get(atomic_number, 0) + 1

    # Form a dictionary that maps Wyckoff letters to a dictionary with the
    # number of atoms of that Wyckoff letter for each atomic number
    wyc_dict: dict = {}
    for i, wyckoff_letter in enumerate(wyckoff_letters):
        old_val = wyc_dict.get(wyckoff_letter, {})
        atomic_number = atomic_numbers[i]
        old_val[atomic_number] = old_val.get(atomic_number, 0) + 1
        wyc_dict[wyckoff_letter] = old_val
    sorted_wyckoff_letters = list(wyc_dict.keys())
    sorted_wyckoff_letters.sort()

    # Anonymize the atomic species to X_<index>, where the index is calculated
    # by ordering the species.
    def compare_atomic_number(at1, at2):
        def cmpp(a, b):
            return ((a < b) - (a > b))

        c = cmpp(atom_count[at1], atom_count[at2])
        if (c != 0):
            return c
        for wyckoff_letter in sorted_wyckoff_letters:
            p = wyc_dict[wyckoff_letter]
            c = cmpp(p.get(at1, 0), p.get(at2, 0))
            if c != 0:
                return c
        return 0
    sorted_species = list(atom_count.keys())
    sorted_species.sort(key=functools.cmp_to_key(compare_atomic_number))
    standard_atom_names = {}
    for i, at in enumerate(sorted_species):
        standard_atom_names[at] = ("X_%d" % i)

    # Rename with anonymized species labels
    standard_wyc: dict = {}
    for wk, ats in wyc_dict.items():
        std_ats = {}
        for at, count in ats.items():
            std_ats[standard_atom_names[at]] = count
        standard_wyc[wk] = std_ats

    # Divide atom counts with greatest common divisor
    if standard_wyc:
        counts = [c for x in standard_wyc.values() for c in x.values()]
        gcd = counts[0]
        for c in counts[1:]:
            gcd = fractions.gcd(gcd, c)
        if gcd != 1:
            for d in standard_wyc.values():
                for at, c in d.items():
                    d[at] = c // gcd

    return standard_wyc


def search_aflow_prototype(space_group: int, norm_wyckoff: dict) -> dict:
    """Searches the AFLOW prototype library for a match for the given space
    group and normalized Wyckoff sequence. The normalized Wyckoff sequence is
    assumed to come from the MatID symmetry routine.

    Currently only contains Part I of the prototype library (M. J. Mehl, D.
    Hicks, C. Toher, O. Levy, R. M. Hanson, G. L. W. Hart, and S. Curtarolo,
    The AFLOW Library of Crystallographic Prototypes: Part 1, Comp. Mat. Sci.
    136, S1-S828 (2017), 10.1016/j.commatsci.2017.01.017)

    Args:
        space_group_number: Space group number
        norm_wyckoff: Normalized Wyckoff occupations

    Returns:
        Dictionary containing the AFLOW prototype information.
    """
    structure_type_info = None
    type_descriptions: Any = aflow_prototypes["prototypes_by_spacegroup"].get(space_group, [])
    for type_description in type_descriptions:
        current_norm_wyckoffs = type_description.get("normalized_wyckoff_matid")
        if current_norm_wyckoffs and current_norm_wyckoffs == norm_wyckoff:
            structure_type_info = type_description
            break
    return structure_type_info


def get_brillouin_zone(reciprocal_lattice: np.array) -> dict:
    """Calculates the Brillouin Zone information from the given reciprocal
    lattice.

    This function uses the crystallographic definition, so there is no factor
    of 2*Pi.

    Args:
        primitive_lattice: The primitive cell as a matrix where rows are the
            cell basis vectors.

    Returns:
        A dictionary containing:
        "vertices": The vertices of the first Brillouin zone
        "faces": The indices of the vertices that make up the faces on the
            first Brillouin zone. The order of these indices matter, because
            only when combined sequentially they form the correct face.
    """
    # Create the near lattice points that surround the origin
    b1 = reciprocal_lattice[0, :]
    b2 = reciprocal_lattice[1, :]
    b3 = reciprocal_lattice[2, :]

    list_k_points = []
    for i, j, k in itertools.product([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]):
        list_k_points.append(i * b1 + j * b2 + k * b3)

    # Create the first Brillouin zone by calculating a Voronoi cell starting
    # from the reciprocal cell origin.
    voronoi = Voronoi(list_k_points)
    origin_index = 13

    # Get the vertices. The regions attribute will contain a list of
    # different regions that were found during the Voronoi creation. We want
    # the Voronoi region for the point at the origin.
    point_region = voronoi.point_region[13]
    vertice_indices = voronoi.regions[point_region]
    vertices = voronoi.vertices[vertice_indices].tolist()

    # Create a mapping between the original index and an index in the new list
    index_map = {
        old_id: new_id for (new_id, old_id) in enumerate(vertice_indices)
    }

    # The ridges are the faces of a 3D Voronoi cell. Here we search for ridges
    # that are placed between the origin and some other point. These form the
    # BZ faces.
    faces = []
    for key in voronoi.ridge_dict:
        if key[0] == origin_index or key[1] == origin_index:
            ridge_indices = voronoi.ridge_dict[key]
            new_ridge_indices = [index_map[i] for i in ridge_indices]
            faces.append(new_ridge_indices)
    faces = faces

    brillouin_zone = {
        "vertices": vertices,
        "faces": faces,
    }
    return brillouin_zone
